"""
abstracted classes, more like interfaces
"""

from typing import Optional

# ==============================================================================

from ..image_wrapper import BasicImage
from ..resizer import ResizerConfig
from ..resizer import ResizerResultset
from ..resizer import TYPE_selected_resizes
from ..utils import TYPE_files_mapping

# ------------------------------------------------------------------------------


class SaverConfig(object):
    archive_original: Optional[bool] = None


class SaverLogger(object):
    def log_save(self, **kwargs) -> None:
        pass

    def log_delete(self, **kwargs) -> None:
        pass


class SaverManagerFactory(object):
    _resizerConfig: ResizerConfig
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger

    def manager(self) -> "SaverManager":  # type: ignore[empty-body]
        pass

    def simple_access(self) -> "SaverSimpleAccess":  # type: ignore[empty-body]
        pass


class _SaverCoreManager(object):
    _resizerConfig: ResizerConfig
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger

    def files_delete(  # type: ignore[empty-body]
        self,
        files_saved: TYPE_files_mapping,
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        """
        workhorse for deletion

        `files_saved`
            `dict`
            format =
                files_saved[size] = (target_filename, bucket_name)
        """
        pass


class SaverManager(_SaverCoreManager):
    def generate_filenames(  # type: ignore[empty-body]
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
    ) -> TYPE_files_mapping:
        pass

    def files_save(  # type: ignore[empty-body]
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
        dry_run: Optional[bool] = False,
    ) -> TYPE_files_mapping:
        pass

    # _SaverCoreManager.files_delete


class SaverSimpleAccess(_SaverCoreManager):
    def file_save(  # type: ignore[empty-body]
        self,
        bucket_name: str,
        filename: str,
        wrappedFile: BasicImage,
        upload_type: str = "public",
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        pass

    def simple_saves_mapping(  # type: ignore[empty-body]
        self,
        bucket_name: str,
        filename: str,
    ) -> TYPE_files_mapping:
        pass

    # _SaverCoreManager.files_delete



# stlib
from typing import Optional

# local
from .. import utils
from ..image_wrapper import ResizerInstructions
from ..resizer import ResizerResultset

# ==============================================================================


def check_archive_original(
    resizerResultset: ResizerResultset,
    archive_original: Optional[bool] = None,
) -> bool:
    """do we want to archive the original?

    `resizerResultset`
        object of `resizer.Resultset`

    `archive_original`
        should we archive original?
        `None` (default)
            implicit
            archive if resizerResultset.original
        `True`
            explict.
            archive resizerResultset.original;
            raise error if missing
        `False`
            explicit.
            do not archive.
    """

    if archive_original is False:
        return False

    elif archive_original is None:
        if resizerResultset.original:
            return True
        return False

    # elif archive_original is True:
    else:
        if not resizerResultset.original:
            raise ValueError(
                """Missing resizerResultset.original for
                explicit archiving"""
            )
        return True


def derive_resized_format(
    size: str,
    resizerResultset: ResizerResultset,
    instructions: ResizerInstructions,
) -> str:
    """derives the format for a size. this function is needed because
    introspection is necessary when the format will be inherited from the
    original file

    args:
        `size` - string
        `resizerResultset` -
        `instructions` - schema dictionary

    ONLY USED BY `size_to_filename`

    SEE ALSO: imagehelper.utils.derive_format

    """
    intended_format: str = instructions["format"]
    if intended_format.upper() in ("AUTO", "ORIGINAL"):
        assert resizerResultset.original.format
        intended_format = utils.derive_format(
            intended_format, resizerResultset.original.format
        )
    return intended_format


def size_to_filename(
    guid: str,
    size: str,
    resizerResultset: ResizerResultset,
    filename_template_default: str,
    instructions: ResizerInstructions,
) -> str:
    """generates the target_filename for a size.

    args:
        `size` - string
        `resizerResultset` -
        `filename_template_default` - string template
        `instructions` - schema dictionary
    """

    # calc vars for filename templating
    filename_template = filename_template_default
    suffix = size
    if "filename_template" in instructions:
        filename_template = instructions["filename_template"]
    if "suffix" in instructions:
        suffix = instructions["suffix"]

    # use a helper to get the format
    # if we used a FakeResultSet to generate the size_to_filename...
    # ... then the resized will be a string of the suffix
    _format = derive_resized_format(size, resizerResultset, instructions)

    # generate the filename
    target_filename = filename_template % {
        "guid": guid,
        "suffix": suffix,
        "format": utils.PIL_type_to_standardized(_format),
    }
    return target_filename



# stdlib
import unittest
import pdb  # noqa
import os

# pypi
import requests

# local
import imagehelper
from imagehelper import _io

# by default, do not test S3 connectivity, as that relies on secrets
TEST_S3 = int(os.environ.get("TEST_S3", 0))

# local - popular
# github secrets
AWS_KEY_PUBLIC = os.environ.get("AWS_KEY_PUBLIC", "12345")
AWS_KEY_SECRET = os.environ.get("AWS_KEY_SECRET", "54321")
AWS_BUCKET_PUBLIC = os.environ.get("AWS_BUCKET_PUBLIC", "bucket-public")
AWS_BUCKET_ARCHIVE = os.environ.get("AWS_BUCKET_ARCHIVE", "bucket-archive")
AWS_BUCKET_ALT = os.environ.get("AWS_BUCKET_ALT", "bucket-alt")
# or export for your test
"""
export TEST_S3=1
export AWS_KEY_PUBLIC=AWS_KEY_PUBLIC
export AWS_KEY_SECRET=AWS_KEY_SECRET
export AWS_BUCKET_PUBLIC=bucket-public
export AWS_BUCKET_ARCHIVE=bucket-archive
export AWS_BUCKET_ALT=bucket-alt
"""


LOCALFILE_DIRECTORY = "tests/localfile-output"


# ------------------------------------------------------------------------------


resizesSchema = {
    "thumb1": {
        "width": 120,
        "height": 120,
        "save_quality": 50,
        "suffix": "t1",
        "format": "JPEG",
        "constraint-method": "fit-within",
        "filename_template": "%(guid)s.%(format)s",
        "boto3_ExtraArgs": {"ACL": "public-read"},
    },
    "t2": {
        "width": 120,
        "height": 120,
        "save_quality": 50,
        "suffix": "t2",
        "format": "PDF",
        "constraint-method": "fit-within:ensure-width",
    },
    "thumb3": {
        "width": 120,
        "height": 120,
        "format": "GIF",
        "constraint-method": "fit-within:ensure-height",
    },
    "t4": {
        "width": 120,
        "height": 120,
        "save_optimize": True,
        "filename_template": "%(guid)s---%(suffix)s.%(format)s",
        "suffix": "t4",
        "format": "PNG",
        "constraint-method": "fit-within:crop-to",
    },
    "t5": {
        "width": 120,
        "height": 120,
        "save_optimize": True,
        "filename_template": "%(guid)s---%(suffix)s.%(format)s",
        "suffix": "t4",
        "format": "auto",
        "constraint-method": "fit-within:crop-to",
    },
}
selected_resizes = ["thumb1", "t2", "thumb3", "t4", "t5"]

# the original is a `gif`
# so we wnat to ensure we safe to a "PNG"
FAKED_resizes = {
    "@archive": ("123456789.gif", "archive"),
    "t2": ("123456789-t2.pdf", "public"),
    "t4": ("123456789---t4.png", "public"),
    "t5": ("123456789---t4.png", "public"),
    "thumb1": ("123456789.jpg", "public"),
    "thumb3": ("123456789-thumb3.gif", "public"),
}


_img = None


def get_imagefile():
    global _img
    if _img is None:
        # py3-  r= _io.TextIOWrapper
        # py3-  rb= _io.BufferedReader
        img = open("tests/test-data/henry.jpg", _io.FileReadArgs)
        img.seek(0)
        data = img.read()
        img.close()
        imgMemory = _io._DefaultMemoryType()
        imgMemory.write(data)
        _img = imgMemory
    _img.seek(0)
    return _img


def newSaverConfig():
    """
    save the files into AmazonS3
    """
    saverConfig = imagehelper.saver.s3.SaverConfig(
        key_public=AWS_KEY_PUBLIC,
        key_private=AWS_KEY_SECRET,
        bucket_public_name=AWS_BUCKET_PUBLIC,
        bucket_archive_name=AWS_BUCKET_ARCHIVE,
        boto3_ExtraArgs_default_public={"ACL": "public-read"},
        boto3_ExtraArgs_default_archive={},
        archive_original=True,
    )

    return saverConfig


def newSaverConfig_Localfile():
    """
    save the files into a folder with the same names as our AWS buckets
    """
    saverConfig = imagehelper.saver.localfile.SaverConfig(
        # subdir_public_name = "directory-public",
        # subdir_archive_name = "directory-archive",
        archive_original=True,
        filedir=LOCALFILE_DIRECTORY,
    )

    return saverConfig


def newResizerConfig(optimize_original=True, optimize_resized=True):
    resizerConfig = imagehelper.resizer.ResizerConfig(
        resizesSchema=resizesSchema,
        selected_resizes=selected_resizes,
        optimize_original=optimize_original,
        optimize_resized=optimize_resized,
    )
    return resizerConfig


def newSaverLogger():
    saverLogger = imagehelper.saver.s3.SaverLogger()
    return saverLogger


class CustomSaverLogger(imagehelper.saver.s3.SaverLogger):
    def __init__(self):
        self._saves = []
        self._deletes = []

    def log_save(self, bucket_name=None, key=None, file_size=None, file_md5=None):
        # todo: change to logger and/or mock
        print("CustomSaverLogger.log_save")
        print("\t %s, %s, %s, %s" % (bucket_name, key, file_size, file_md5))
        self._saves.append((bucket_name, key, file_size, file_md5))

    def log_delete(self, bucket_name=None, key=None):
        # todo: change to logger and/or mock
        print("CustomSaverLogger.log_delete")
        print("\t %s, %s" % (bucket_name, key))
        self._deletes.append((bucket_name, key))


class _ImagehelperTestingMixin(object):
    def _check_resizedImages(self, resizedImages):
        # ensure the original has a file size
        self.assertNotEqual(resizedImages.original.file_size, 0)

        # ensure every resize has a file_size
        for _size in resizedImages.resized.keys():
            self.assertNotEqual(resizedImages.resized[_size].file_size, 0)

    def _check_fakeResizedImages(self, resizedImages):
        # ensure we got the faked version
        self.assertIsInstance(
            resizedImages.original, imagehelper.image_wrapper.FakedOriginal
        )

        # ensure we got the faked version
        for _size in resizedImages.resized.keys():
            self.assertIsInstance(
                resizedImages.resized[_size], imagehelper.image_wrapper.FakedResize
            )

    def _build_faked(self):
        # new resizer config
        resizerConfig = newResizerConfig()

        # build a new resizer
        resizer = imagehelper.resizer.Resizer(resizerConfig=resizerConfig)

        # resize the image
        resizedImages = resizer.fake_resize(original_filename="300x200.gif")

        return resizerConfig, resizer, resizedImages


class TestResize(unittest.TestCase, _ImagehelperTestingMixin):
    def test_direct_resize(self):
        # new resizer config
        resizerConfig = newResizerConfig()

        # build a new resizer
        resizer = imagehelper.resizer.Resizer(resizerConfig=resizerConfig)

        # try to register the image
        resizer.register_image_file(imagefile=get_imagefile())

        try:
            # resize the image
            # this should fail, because we don't want to risk changing the image before registering
            results = resizer.resize(imagefile=get_imagefile())
            raise ValueError("this should have failed")
        except imagehelper.errors.ImageError_DuplicateAction:
            # expected!
            pass

        # build a new resizer
        resizer = imagehelper.resizer.Resizer(resizerConfig=resizerConfig)

        # resize the image
        resizedImages = resizer.resize(imagefile=get_imagefile())

        # audit the payload
        self._check_resizedImages(resizedImages)

    def test_fake_resize(self):
        resizerConfig, resizer, resizedImages = self._build_faked()
        # audit the payload
        self._check_fakeResizedImages(resizedImages)


class TestS3(unittest.TestCase, _ImagehelperTestingMixin):
    def test_s3_factory(self):
        # generate the configs
        resizerConfig = newResizerConfig()
        saverConfig = newSaverConfig()
        saverLogger = newSaverLogger()

        # generate the factory
        saverManagerFactory = imagehelper.saver.s3.SaverManagerFactory(
            saverConfig=saverConfig,
            saverLogger=saverLogger,
            resizerConfig=resizerConfig,
        )

        # grab a manager
        saverManager = saverManagerFactory.manager()

        # make sure we generated a manager
        assert isinstance(saverManager, imagehelper.saver.s3.SaverManager)

        # inspect the manager to ensure it is set up correctly
        assert saverManager._saverConfig == saverConfig
        assert saverManager._saverLogger == saverLogger
        assert saverManager._resizerConfig == resizerConfig

    @unittest.skipUnless(TEST_S3, "S3 Testing Disabled")
    def test_s3__saver_manager(self):
        """
        test saving files with the `localfile.SaverManager`
        """

        # new resizer config
        resizerConfig = newResizerConfig()
        # build a factory
        resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)

        # grab a resizer
        resizer = resizerFactory.resizer()

        # resize !
        resizedImages = resizer.resize(imagefile=get_imagefile())

        # audit the payload
        self._check_resizedImages(resizedImages)

        # new s3 config
        saverConfig = newSaverConfig()
        # new s3 logger
        # saverLogger = imagehelper.saver.s3.SaverLogger()
        saverLogger = CustomSaverLogger()

        # upload the resized items
        uploader = imagehelper.saver.s3.SaverManager(
            saverConfig=saverConfig,
            resizerConfig=resizerConfig,
            saverLogger=saverLogger,
        )

        guid = "123"
        uploaded = uploader.files_save(resizedImages, guid)
        deleted = uploader.files_delete(uploaded)

    @unittest.skipUnless(TEST_S3, "S3 Testing Disabled")
    def test_s3__saver_simple_access(self):
        """
        test saving files with the `localfile.SaverSimpleAccess`
        """

        # new resizer config
        resizerConfig = newResizerConfig()
        # build a factory
        resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)

        # grab a resizer
        resizer = resizerFactory.resizer()

        # resize !
        resizedImages = resizer.resize(imagefile=get_imagefile())

        # audit the payload
        self._check_resizedImages(resizedImages)

        # new s3 config
        saverConfig = newSaverConfig()
        # new s3 logger
        # saverLogger = imagehelper.saver.s3.SaverLogger()
        saverLogger = CustomSaverLogger()

        # upload the resized items
        uploader = imagehelper.saver.s3.SaverSimpleAccess(
            saverConfig=saverConfig,
            resizerConfig=resizerConfig,
            saverLogger=saverLogger,
        )

        guid = "123"
        filename = "%s.jpg" % guid
        bucket_name = saverConfig.bucket_public_name
        uploaded = uploader.file_save(
            bucket_name,
            filename,
            resizedImages.resized["thumb1"],
            upload_type="public",
            dry_run=False,
        )
        assert uploaded
        assert len(saverLogger._saves) == 1

        s3_url = "https://%s.s3.amazonaws.com/%s" % (bucket_name, filename)
        resp = requests.get(s3_url)
        assert resp.status_code == 200
        assert resp.headers["Content-Type"] == "image/jpeg"

        # but still rely on the parent class' call `files_delete` (fileS)
        deleted = uploader.files_delete(uploaded)
        assert isinstance(deleted, dict)
        assert len(deleted.items()) == 0
        assert len(saverLogger._deletes) == 1

        # and check it's not available any longer
        resp2 = requests.get(s3_url)
        assert resp2.status_code == 403


class TestLocalfile(unittest.TestCase, _ImagehelperTestingMixin):
    def test_localfile__saver_manager(self):
        """
        test saving files with the `localfile.SaverManager`
        """

        # new resizer config
        resizerConfig = newResizerConfig()
        # build a factory
        resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)

        # grab a resizer
        resizer = resizerFactory.resizer()

        # resize !
        resizedImages = resizer.resize(imagefile=get_imagefile())

        # audit the payload
        self._check_resizedImages(resizedImages)

        # new s3 config
        saverConfig = newSaverConfig_Localfile()

        # new s3 logger
        saverLogger = imagehelper.saver.localfile.SaverLogger()

        # upload the resized items
        saver = imagehelper.saver.localfile.SaverManager(
            saverConfig=saverConfig,
            resizerConfig=resizerConfig,
            saverLogger=saverLogger,
        )

        guid = "123"
        uploaded = saver.files_save(resizedImages, guid)
        deleted = saver.files_delete(uploaded)

    maxDiff = None

    def test_localfile__saver_manager__faked(self):
        resizerConfig, resizer, resizedImages = self._build_faked()
        # audit the payload
        self._check_fakeResizedImages(resizedImages)

        # new s3 config
        saverConfig = newSaverConfig_Localfile()

        # new s3 logger
        saverLogger = imagehelper.saver.localfile.SaverLogger()

        # upload the resized items
        saver = imagehelper.saver.localfile.SaverManager(
            saverConfig=saverConfig,
            resizerConfig=resizerConfig,
            saverLogger=saverLogger,
        )

        guid = "123456789"

        # generate_filenames
        fnames = saver.generate_filenames(
            resizedImages,
            guid,
            selected_resizes=selected_resizes,
            archive_original=True,
        )

        self.assertEqual(fnames, FAKED_resizes)

    def test_localfile__saver_simple_access(self):
        """
        test saving files with the `localfile.SaverSimpleAccess`
        """

        # new resizer config
        resizerConfig = newResizerConfig()
        # build a factory
        resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)

        # grab a resizer
        resizer = resizerFactory.resizer()

        # resize !
        resizedImages = resizer.resize(imagefile=get_imagefile())

        # audit the payload
        self._check_resizedImages(resizedImages)

        # new saver config
        saverConfig = newSaverConfig_Localfile()
        # new logger
        saverLogger = imagehelper.saver.localfile.SaverLogger()

        # upload the resized items
        saver = imagehelper.saver.localfile.SaverSimpleAccess(
            saverConfig=saverConfig,
            resizerConfig=resizerConfig,
            saverLogger=saverLogger,
        )

        subdir_name = "test_localfile_simple"
        filename = "thumb1.jpg"

        # note we call `file_save` (file)
        uploaded = saver.file_save(
            subdir_name, filename, resizedImages.resized["thumb1"]
        )
        assert uploaded

        # test the directory is there
        _subdirs = os.listdir(LOCALFILE_DIRECTORY)
        assert "test_localfile_simple" in _subdirs

        # but still rely on the parent class' call `files_delete` (fileS)
        deleted = saver.files_delete(uploaded)

        # test the directory is now not-there, because it was cleaned up
        _subdirs = os.listdir(LOCALFILE_DIRECTORY)
        assert "test_localfile_simple" not in _subdirs


class TestResizingMethods(unittest.TestCase, _ImagehelperTestingMixin):
    def test_fit_within(self):
        method = "fit-within"
        schema = {
            "test": {
                "width": 120,
                "height": 120,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (90, 120)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_crop_to(self):
        method = "fit-within:crop-to"
        schema = {
            "test": {
                "width": 120,
                "height": 120,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (120, 120)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_ensure_width(self):
        method = "fit-within:ensure-width"
        schema = {
            "test": {
                "width": 120,
                "height": 120,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (120, 160)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_ensure_height(self):
        method = "fit-within:ensure-height"
        schema = {
            "test": {
                "width": 120,
                "height": 120,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (90, 120)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )
        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_smallest_ensure_minimum(self):
        method = "smallest:ensure-minimum"
        schema = {
            "test": {
                "width": 120,
                "height": 120,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (120, 160)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_exact_no_resize(self):
        method = "exact:no-resize"
        schema = {
            "test": {
                "width": 1200,
                "height": 1600,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (1200, 1600)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]

    def test_fit_within_exact_proportion(self):
        method = "exact:proportion"
        schema = {
            "test": {
                "width": 240,
                "height": 320,
                "save_optimize": True,
                "format": "PNG",
                "constraint-method": method,
            }
        }

        # what do we expect ?
        expected_original_wh = (1200, 1600)
        expected_resized_wh = (240, 320)

        r = imagehelper.resizer.Resizer()
        resizedImages = r.resize(
            imagefile=get_imagefile(),
            resizesSchema=schema,
            selected_resizes=("test",),
            optimize_original=False,
            optimize_resized=True,
        )

        # audit the payload
        self._check_resizedImages(resizedImages)

        # what do we have ?
        actual_original_wh = (
            resizedImages.original.width,
            resizedImages.original.height,
        )
        actual_resized_wh = (
            resizedImages.resized["test"].width,
            resizedImages.resized["test"].height,
        )

        # assert the original matches
        assert expected_original_wh[0] == actual_original_wh[0]
        assert expected_original_wh[1] == actual_original_wh[1]

        # assert the resize matches
        assert expected_resized_wh[0] == actual_resized_wh[0]
        assert expected_resized_wh[1] == actual_resized_wh[1]



from __future__ import print_function

# stdlib
import os
import pprint

# local
import imagehelper
from imagehelper import _io

# ------------------------------------------------------------------------------

# retrieve credentials/configs from the environment
AWS_KEY_PUBLIC = os.environ.get("AWS_KEY_PUBLIC")
AWS_KEY_SECRET = os.environ.get("AWS_KEY_SECRET")
AWS_BUCKET_PUBLIC = os.environ.get("AWS_BUCKET_PUBLIC")
AWS_BUCKET_ARCHIVE = os.environ.get("AWS_BUCKET_ARCHIVE")
AWS_BUCKET_ALT = os.environ.get("AWS_BUCKET_ALT")

# use the credentials to generate a S3Saver configuration
saverConfig = imagehelper.saver.s3.SaverConfig(
    key_public=AWS_KEY_PUBLIC,
    key_private=AWS_KEY_SECRET,
    bucket_public_name=AWS_BUCKET_PUBLIC,
    bucket_archive_name=AWS_BUCKET_ARCHIVE,
    boto3_ExtraArgs_default_public={"ACL": "public-read"},
    boto3_ExtraArgs_default_archive={},
    archive_original=True,
)

# define a schema for resizing
resizesSchema = {
    "thumb1": {
        "width": 120,
        "height": 120,
        "save_quality": 50,
        "suffix": "t1",
        "format": "JPEG",
        "constraint-method": "fit-within",
        "s3_bucket_public": AWS_BUCKET_ALT,
        "filename_template": "%(guid)s.%(format)s",
        "boto3_ExtraArgs": {"ACL": "public-read"},
    },
    "t2": {
        "width": 120,
        "height": 120,
        "save_quality": 50,
        "suffix": "t2",
        "format": "PDF",
        "constraint-method": "fit-within:ensure-width",
    },
    "thumb3": {
        "width": 120,
        "height": 120,
        "format": "GIF",
        "constraint-method": "fit-within:ensure-height",
    },
    "t4": {
        "width": 120,
        "height": 120,
        "save_optimize": True,
        "filename_template": "%(guid)s---%(suffix)s.%(format)s",
        "suffix": "t4",
        "format": "PNG",
        "constraint-method": "fit-within:crop-to",
    },
}

# specify the keys to resize
selected_resizes = ["thumb1", "t2", "thumb3", "t4"]

# specify an alternate schema for resizing
resizesSchema_alt = {
    "og:image": {
        "width": 3200,
        "height": 3200,
        "save_quality": 50,
        "suffix": "og",
        "format": "JPEG",
        "constraint-method": "smallest:ensure-minimum",
        "filename_template": "%(guid)s-og.%(format)s",
    },
    "og:image2": {
        "width": 200,
        "height": 200,
        "save_quality": 50,
        "suffix": "og",
        "format": "JPEG",
        "constraint-method": "smallest:ensure-minimum",
        "filename_template": "%(guid)s-og.%(format)s",
    },
    "og:image3": {
        "width": 300,
        "height": 200,
        "save_quality": 50,
        "suffix": "og",
        "format": "JPEG",
        "constraint-method": "smallest:ensure-minimum",
        "filename_template": "%(guid)s-og.%(format)s",
    },
}


# this will be used for logging S3 interactions
class CustomSaverLogger(imagehelper.saver.s3.SaverLogger):
    def log_save(self, bucket_name=None, key=None, file_size=None, file_md5=None):
        print("CustomSaverLogger.log_save")
        print("\t %s, %s, %s, %s" % (bucket_name, key, file_size, file_md5))

    def log_delete(self, bucket_name=None, key=None):
        print("CustomSaverLogger.log_delete")
        print("\t %s, %s" % (bucket_name, key))


saverLogger = CustomSaverLogger()
resizerConfig = imagehelper.resizer.ResizerConfig(
    resizesSchema=resizesSchema,
    selected_resizes=selected_resizes,
    optimize_original=True,
    optimize_resized=True,
)


# create a factory
resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)


_img = None


def get_imagefile():
    global _img
    if _img is None:
        img = open("tests/test-data/henry.jpg", _io.FileReadArgs)
        img.seek(0)
        data = img.read()
        img.close()
        imgMemory = _io._DefaultMemoryType()
        imgMemory.write(data)
        _img = imgMemory
    _img.seek(0)
    return _img


guid = "123"


def demo_direct():
    "demo calling direct methods"
    print(">>> demo_direct | start")

    resizer = resizerFactory.resizer()

    # try to register the image
    resizer.register_image_file(imagefile=get_imagefile())

    try:
        # resize the image
        # this should fail, because we don't want to risk changing the image before registering
        results = resizer.resize(imagefile=get_imagefile())
    except imagehelper.errors.ImageError_DuplicateAction:
        # expected!
        pass

    resizer = resizerFactory.resizer(imagefile=get_imagefile())
    resizedImages = resizer.resize()

    print(". resized the image. report follows:")
    pprint.pprint(resizedImages.__dict__)
    print("<<< demo_direct | done")


def demo_factory():
    "demo calling factory methods"
    print(">>> demo_factory | start")

    # resize !
    resizer = resizerFactory.resizer(imagefile=get_imagefile())
    resizedImages = resizer.resize()

    if not os.path.exists("demo-output/demo"):
        os.makedirs("demo-output/demo")
    for k in resizedImages.resized.keys():
        _filename = "demo-output/demo/%s.%s" % (k, resizesSchema[k]["format"])
        print(". writing %s" % _filename)
        open(_filename, _io.FileWriteArgs).write(
            resizedImages.resized[k].file.getvalue()
        )

    resizedImages.original.optimize()
    open("demo-output/demo/original.png", _io.FileWriteArgs).write(
        resizedImages.original.file.getvalue()
    )
    print("<<< demo_factory | done")


def demo_s3():
    "demo s3 uploading"
    print(">>> demo_s3 | start")

    resizer = resizerFactory.resizer(imagefile=get_imagefile())
    resizedImages = resizer.resize()

    # upload the resized items
    uploader = imagehelper.saver.s3.SaverManager(
        saverConfig=saverConfig, resizerConfig=resizerConfig, saverLogger=saverLogger
    )
    uploaded = uploader.files_save(resizedImages, guid)
    print(". uploaded! %s" % uploaded)

    deleted = uploader.files_delete(uploaded)
    print(". deleted! %s" % deleted)
    print("<<< demo_s3 | done")


def demo_s3_alt():
    "demo s3 uploading"
    print(">>> demo_s3_alt | start")

    resizer = resizerFactory.resizer(imagefile=get_imagefile())
    resizedImages = resizer.resize()

    # upload the resized items
    uploader = imagehelper.saver.s3.SaverManager(
        saverConfig=saverConfig, resizerConfig=resizerConfig, saverLogger=saverLogger
    )
    uploaded_original = uploader.files_save(
        resizedImages, guid, selected_resizes=[], archive_original=True
    )
    print(". uploaded_original! %s" % uploaded_original)

    uploaded_resizes = uploader.files_save(resizedImages, guid, archive_original=False)
    print(". uploaded_resizes! %s" % uploaded_resizes)

    uploaded_all = dict(
        list(uploaded_original.items()) + list(uploaded_resizes.items())
    )

    deleted = uploader.files_delete(uploaded_all)
    print(". deleted! %s" % deleted)
    print("<<< demo_s3_alt | done")


def _print_ResizedImages(resizedImages):
    print(". resizedImages report: ")
    print(".. original file_md5:", resizedImages.original.file_md5)
    print(".. original file_size:", resizedImages.original.file_size)
    print(".. for the resizes...")
    print(".. resize | file_size |  file_md5")
    for k in resizedImages.resized.keys():
        print(
            ". ",
            k,
            "|",
            resizedImages.resized[k].file_size,
            "|",
            resizedImages.resized[k].file_md5,
        )


def demo_alt_resizing():
    print(">>> demo_alt_resizing | start")

    resizerConfig = imagehelper.resizer.ResizerConfig(
        resizesSchema=resizesSchema_alt,
        optimize_original=True,
        optimize_resized=True,
    )

    # build a factory & resize
    resizerFactory = imagehelper.resizer.ResizerFactory(resizerConfig=resizerConfig)

    # not a big difference here...
    # usually we would call
    # - resizedImages = resizerFactory.resizer(imagefile=get_imagefile()).resize()
    # but instead we are deferring the imagefile into the resize() function
    resizedImages = resizerFactory.resizer().resize(imagefile=get_imagefile())
    for k in resizedImages.resized.keys():
        print(
            ". ",
            k,
            resizedImages.resized[k].file_size,
            resizedImages.resized[k].file_md5,
        )
    _print_ResizedImages(resizedImages)

    print("<<< demo_alt_resizing | done")


def demo_md5():
    "demo file md5"
    print(">>> demo_md5 | start")

    # resize !
    resizer = resizerFactory.resizer(imagefile=get_imagefile())
    resizedImages = resizer.resize()
    _print_ResizedImages(resizedImages)
    print("<<< demo_md5 | done")


def demo_serialze():
    "demo file serialize, this just looks creating a file from a b64 encoding"
    print(">>> demo_serialze | start")

    # resize !
    resizer = resizerFactory.resizer(imagefile=get_imagefile())

    file_md5 = resizer.get_original().file_md5
    file_b64 = resizer.get_original().file_b64

    resizer = resizerFactory.resizer(file_b64=file_b64)
    print("<<< demo_serialze | done")


demo_alt_resizing()
demo_factory()
demo_direct()
demo_md5()

if AWS_KEY_PUBLIC:
    demo_s3()
    demo_s3_alt()
else:
    print("no AWS_KEY_PUBLIC set, not running s3 ")



# stdlb
from io import BufferedReader
import logging
from typing import Dict
from typing import Optional

# from io import RawIOBase

# pypi
try:
    import boto3
except ImportError:
    boto3 = None

# local
from . import _core
from .utils import check_archive_original
from .utils import size_to_filename
from .. import errors
from .. import utils
from ..image_wrapper import BasicImage
from ..image_wrapper import ResizerInstructions
from ..resizer import ResizerConfig
from ..resizer import ResizerResultset
from ..resizer import TYPE_selected_resizes
from ..utils import TYPE_files_mapping

# ==============================================================================


log = logging.getLogger(__name__)


class NoBoto(ImportError):
    pass


NO_BOTO = NoBoto("`boto3` was not available for import")


class ResizerInstructions_S3(ResizerInstructions):
    s3_bucket_public: str
    boto3_ExtraArgs: Dict[str, str]


TYPE_ResizesSchema_S3 = Dict[str, ResizerInstructions_S3]


class ResizerConfig_S3(ResizerConfig):
    resizesSchema: TYPE_ResizesSchema_S3  # type: ignore[assignment]


# ------------------------------------------------------------------------------


# note: dirty workaround for boto3
"""
boto3 has an absurd flaw in which fileobjects are closed when uploaded
(see https://github.com/boto/s3transfer/issues/80)

This is a temporary workaround
"""


class NonCloseableBufferedReader(BufferedReader):

    """
    # PY3 Typing;

    from typing import Any
    from typing import IO
    from typing import cast

    def __init__(self, raw: IO[bytes], *args: Any, **kwargs: Any):
       super().__init__(cast(RawIOBase, raw), *args, **kwargs)

    def close(self) -> None:
       self.flush()

    """

    def __init__(self, raw, *args, **kwargs):
        super().__init__(raw, *args, **kwargs)

    def close(self):
        self.flush()


# ==============================================================================


class SaverConfig(_core.SaverConfig):
    """
    Configuration data for Amazon S3 Services
    """

    key_public: str
    key_private: str
    bucket_public_name: Optional[str] = None
    bucket_archive_name: Optional[str] = None
    boto3_ExtraArgs_default_public: Dict
    boto3_ExtraArgs_default_archive: Dict

    def __init__(
        self,
        key_public: str = "",
        key_private: str = "",
        bucket_public_name: Optional[str] = None,
        bucket_archive_name: Optional[str] = None,
        boto3_ExtraArgs_default_public: Optional[Dict] = None,
        boto3_ExtraArgs_default_archive: Optional[Dict] = None,
        archive_original: Optional[bool] = None,
        **kwargs,
    ):
        self.key_public = key_public
        self.key_private = key_private
        self.bucket_public_name = bucket_public_name
        self.bucket_archive_name = bucket_archive_name
        self.boto3_ExtraArgs_default_public = boto3_ExtraArgs_default_public or {}
        self.boto3_ExtraArgs_default_archive = boto3_ExtraArgs_default_archive or {}
        self.archive_original = archive_original or False

        # v0.6.0 removed this options
        # raise Exceptions to catch incompatibilities; remove in future release
        if ("bucket_public_headers" in kwargs) or ("bucket_archive_headers" in kwargs):
            raise ValueError(
                """`bucket_public_headers` and `bucket_archive_headers` were """
                """removed in v0.6.0. Please update to use """
                """`boto3_ExtraArgs_default_public` and """
                """`boto3_ExtraArgs_default_archive`."""
            )


class SaverLogger(_core.SaverLogger):
    """The s3 save method will log to this logger on uploads and deletes.
    Any object offering these methods can be replaced;
    This is only illustrative."""

    def log_save(  # type: ignore[override]
        self,
        bucket_name: str = "",
        key: str = "",
        file_size: int = 0,
        file_md5: str = "",
    ) -> None:
        """args:
        `self`
        `bucket_name`
            s3 bucket name
        `key`
            key in bucket
        `file_size`
            size in bytes
        `file_md5`
            hexdigest
        """
        pass

    def log_delete(  # type: ignore[override]
        self,
        bucket_name: str = "",
        key: str = "",
    ) -> None:
        """args:
        `self`
        `bucket_name`
            s3 bucket name
        `key`
            key in bucket
        """
        pass


class SaverManagerFactory(_core.SaverManagerFactory):
    """Factory for generating SaverManager instances"""

    _resizerConfig: ResizerConfig_S3
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_S3,
    ):
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig

    def manager(self) -> "SaverManager":
        """generate and return a new SaverManager instance"""
        return SaverManager(
            saverConfig=self._saverConfig,
            saverLogger=self._saverLogger,
            resizerConfig=self._resizerConfig,
        )

    def simple_access(self) -> "SaverSimpleAccess":
        """generate and return a new SaverSimpleAccess instance"""
        return SaverSimpleAccess(
            saverConfig=self._saverConfig,
            saverLogger=self._saverLogger,
            resizerConfig=self._resizerConfig,
        )


class _SaverCoreManager(_core._SaverCoreManager):
    """
    based on interface defined in `_core.SaverCoreManager`
    """

    _resizerConfig: ResizerConfig_S3
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger

    _s3_client: Optional[boto3.client] = None
    _s3_bucketnames: Optional[Dict] = None
    _boto3_ExtraArgs_default_public: Dict[str, str]  # __init__ -> _generate_defaults
    _boto3_ExtraArgs_default_archive: Dict[str, str]  # __init__ -> _generate_defaults

    filename_template: str = "%(guid)s-%(suffix)s.%(format)s"
    filename_template_archive: str = "%(guid)s.%(format)s"

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_S3,
    ):
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig
        self._generate_defaults()

    def _generate_defaults(self) -> None:
        """
        generate the default headers
        """
        assert self._saverConfig

        # public and archive get different acls / content-types
        self._boto3_ExtraArgs_default_public = {"ACL": "public-read"}
        self._boto3_ExtraArgs_default_archive = {}

        if self._saverConfig.boto3_ExtraArgs_default_public:
            for k, v in self._saverConfig.boto3_ExtraArgs_default_public.items():
                self._boto3_ExtraArgs_default_public[k] = v
        if self._saverConfig.boto3_ExtraArgs_default_archive:
            for k, v in self._saverConfig.boto3_ExtraArgs_default_archive.items():
                self._boto3_ExtraArgs_default_archive[k] = v

    @property
    def s3_client(self) -> boto3.client:
        """property that memoizes the connection"""
        assert self._saverConfig

        if self._s3_client is None:
            if boto3 is None:
                raise NO_BOTO
            self._s3_client = boto3.client(
                "s3",
                aws_access_key_id=self._saverConfig.key_public,
                aws_secret_access_key=self._saverConfig.key_private,
            )
        return self._s3_client

    @property
    def s3_bucketnames(self) -> Dict:
        """property that memoizes the calcuated s3 s3_bucketnames"""
        assert self._saverConfig

        if self._s3_bucketnames is None:
            if boto3 is None:
                raise NO_BOTO

            # memoize the buckets

            # create our bucket list
            s3_bucketnames: Dict = {}

            # @public and @archive are special
            bucketname_public = self._saverConfig.bucket_public_name
            bucketname_archive = self._saverConfig.bucket_archive_name

            s3_bucketnames[bucketname_public] = bucketname_public
            s3_bucketnames["@public"] = bucketname_public
            if self._saverConfig.bucket_archive_name:
                s3_bucketnames[bucketname_archive] = bucketname_archive
                s3_bucketnames["@archive"] = bucketname_archive

            # look through our selected sizes
            assert self._resizerConfig
            if self._resizerConfig:
                assert self._resizerConfig.selected_resizes
                for size in self._resizerConfig.selected_resizes:
                    if size[0] == "@":
                        raise errors.ImageError_ConfigError(
                            "@ is a reserved initial character for image sizes"
                        )

                    if "s3_bucket_public" in self._resizerConfig.resizesSchema[size]:
                        bucket_name = self._resizerConfig.resizesSchema[size][
                            "s3_bucket_public"
                        ]
                        if bucket_name not in s3_bucketnames:
                            s3_bucketnames[bucket_name] = bucket_name

            # store the buckets
            self._s3_bucketnames = s3_bucketnames

        # return the memoized buckets
        return self._s3_bucketnames

    def files_delete(
        self,
        files_saved: TYPE_files_mapping,
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        """workhorse for deletion

        `files_saved`
            `dict`
            format =
                files_saved[size] = (target_filename, bucket_name)

        `dry_run`
            default = `False`
            should we just pretend to save?
        """

        # preload the calculated s3 bucketnames
        s3_bucketnames = self.s3_bucketnames

        # convert to a list, because we delete the items from the dict
        for size in list(files_saved.keys()):
            # grab the stash
            (target_filename, _bucket_name) = files_saved[size]

            # get the calculated bucket_name
            bucket_name = s3_bucketnames[_bucket_name]

            # delete it
            log.debug("going to delete `%s` from `%s`" % (target_filename, bucket_name))

            if not dry_run:
                _del_dicts = [
                    {"Key": target_filename},
                ]
                response = self.s3_client.delete_objects(  # noqa: F841
                    Bucket=bucket_name,
                    Delete={"Objects": _del_dicts},
                )

                # external logging
                if self._saverLogger:
                    self._saverLogger.log_delete(
                        bucket_name=bucket_name, key=target_filename
                    )

            # internal cleanup
            del files_saved[size]

        return files_saved


class SaverManager(_SaverCoreManager, _core.SaverManager):
    """
    `SaverManager` handles all the actual uploading and deleting

    based on interface defined in `_core._SaverCoreManager`
    but inherits from this file's '`_SaverCoreManager`
    """

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_S3,
    ):
        super(SaverManager, self).__init__(
            saverConfig=saverConfig,
            saverLogger=saverLogger,
            resizerConfig=resizerConfig,
        )
        if not resizerConfig:
            raise ValueError(
                """`SaverManager` requires a `resizerConfig` which contains the resize recipes. these are needed for generating filenames."""
            )

    def _validate__selected_resizes(
        self,
        resizerResultset: ResizerResultset,
        selected_resizes: Optional[TYPE_selected_resizes],
    ) -> TYPE_selected_resizes:
        """shared validation
            returns `dict` selected_resizes

        ARGS
            `resizerResultset`
                `resizer.ResizerResultset` object
                    `resized` - dict of images that were resized
                    `original ` - original file

            `selected_resizes`
                iterable of selected resizes

        """

        # default to the resized images
        if selected_resizes is None:
            selected_resizes = list(resizerResultset.resized.keys())

        for k in selected_resizes:
            if k not in resizerResultset.resized:
                raise errors.ImageError_ConfigError(
                    "selected size is not resizerResultset.resized (`%s`)" % k
                )

            assert self._resizerConfig
            if k not in self._resizerConfig.resizesSchema:
                raise errors.ImageError_ConfigError(
                    "selected size is not self._resizerConfig.resizesSchema (`%s`)" % k
                )

            # exist early for invalid sizes
            if k[0] == "@":
                raise errors.ImageError_ConfigError(
                    "@ is a reserved initial character for image sizes (`%s`)" % k
                )

        return selected_resizes

    def generate_filenames(
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
    ) -> TYPE_files_mapping:
        """
        generates the filenames s3 would save to;
        this is useful for planning/testing or deleting old files

        Returns a `dict` of target filenames
            keys = resized size
            values = tuple (target_filename, bucket_name)

        `resizerResultset`
            a `resizer.ResizerResultset` object
                `resized` - dict of images that were resized
                `original ` - original file

        `guid`
            a `uuid` or similar name that forms the basis for storage
            the guid is passed into the template in self._resizerConfig

        `selected_resizes`
            default = `None` -- all keys of resizerResultset.resized
            a `list` of keys to save
            we default to saving all the resized images

        `archive_original`
            default = `None`
            should we archive the original ?
            implicit/explicit archival option.  see `def check_archive_original`

        """
        if guid is None:
            raise errors.ImageError_ArgsError(
                """You must supply a `guid` for
            the image. this is used for filename templating"""
            )

        # default to the resized images
        if selected_resizes is None:
            selected_resizes = list(resizerResultset.resized.keys())

        # quickly validate
        selected_resizes = self._validate__selected_resizes(
            resizerResultset, selected_resizes
        )

        # init our return dict
        filename_mapping: TYPE_files_mapping = {}

        assert self._resizerConfig
        assert self._saverConfig
        for size in selected_resizes:
            instructions = self._resizerConfig.resizesSchema[size]
            target_filename = size_to_filename(
                guid, size, resizerResultset, self.filename_template, instructions
            )

            # figure out the bucket_name
            bucket_name = self._saverConfig.bucket_public_name
            if "s3_bucket_public" in instructions:
                bucket_name = instructions["s3_bucket_public"]
            if not bucket_name:
                raise ValueError("Could not detect a bucket_name for `%s`" % size)
            filename_mapping[size] = (target_filename, bucket_name)

        if check_archive_original(resizerResultset, archive_original=archive_original):
            assert resizerResultset.original.format
            filename_template_archive = self.filename_template_archive
            target_filename = filename_template_archive % {
                "guid": guid,
                "format": utils.PIL_type_to_standardized(
                    resizerResultset.original.format
                ),
            }
            bucket_name = self._saverConfig.bucket_archive_name
            if not bucket_name:
                raise ValueError("Could not detect a bucket_name for archive")
            filename_mapping["@archive"] = (target_filename, bucket_name)

        # return the filemapping
        return filename_mapping

    def files_save(  # type: ignore[override]
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        """
        Returns a dict of resized images
        calls self.register_image_file() if needed

        this resizes the images.
        it returns the images and updates the internal dict.

        `resizerResultset`
            a `resizer.ResizerResultset` object
                `resized` - dict of images that were resized
                `original ` - original file

        `guid`
            a `uuid` or similar name that forms the basis for storage
            the guid is passed to the template in `self.generate_filenames`

        `selected_resizes`
            default = `None` -- all keys of resizerResultset.resized
            a `list` of keys to save
            we default to saving all the resized images
            if you don't want to resize any images:
                pass in an empty list -- []
                passing in `None` will run the default images

        `archive_original`
            default = `None`
            should we archive the original ?
            implicit/explicit archival option.  see `def check_archive_original`

        `dry_run`
            default = `False`
            should we just pretend to upload
        """
        if boto3 is None:
            raise NO_BOTO

        if guid is None:
            raise errors.ImageError_ArgsError(
                """You must supply a `guid` for
            the image. this is used"""
            )

        # quickly validate
        selected_resizes = self._validate__selected_resizes(
            resizerResultset, selected_resizes
        )

        # preload the calculated s3 bucketnames
        s3_bucketnames = self.s3_bucketnames

        # and then we have the bucketed filenames...
        target_filenames = self.generate_filenames(
            resizerResultset,
            guid,
            selected_resizes=selected_resizes,
            archive_original=archive_original,
        )

        # log uploads for removal/tracking and return
        files_saved: TYPE_files_mapping = {}
        assert self._resizerConfig
        assert self._resizerConfig.resizesSchema
        try:
            # and then we upload...
            for size in selected_resizes:
                # grab the stash
                (target_filename, _bucket_name) = target_filenames[size]

                # get the calculated bucket_name
                bucket_name = s3_bucketnames[_bucket_name]

                log.debug("Uploading `%s` to `%s` " % (target_filename, bucket_name))

                # generate the ExtraArgs
                _boto3_ExtraArgs = self._boto3_ExtraArgs_default_public.copy()
                _format = resizerResultset.resized[size].format
                assert isinstance(_format, str)
                _boto3_ExtraArgs["ContentType"] = utils.PIL_type_to_content_type(
                    _format
                )
                # overwrite with Resizer ExtraArgs
                if "boto3_ExtraArgs" in self._resizerConfig.resizesSchema[size]:
                    for k in self._resizerConfig.resizesSchema[size]["boto3_ExtraArgs"]:
                        _boto3_ExtraArgs[k] = self._resizerConfig.resizesSchema[size][
                            "boto3_ExtraArgs"
                        ][k]

                if not dry_run:
                    # the active file
                    _wrapped = resizerResultset.resized[size]

                    # dirty workaround. boto3 has a bug in which it closes files.
                    _buffer = NonCloseableBufferedReader(_wrapped.file)

                    # upload
                    response = self.s3_client.upload_fileobj(
                        _buffer,
                        bucket_name,
                        target_filename,
                        ExtraArgs=_boto3_ExtraArgs,
                    )
                    _buffer.detach()

                    # log to external plugin too
                    if self._saverLogger:
                        self._saverLogger.log_save(
                            bucket_name=bucket_name,
                            key=target_filename,
                            file_size=_wrapped.file_size,
                            file_md5=_wrapped.file_md5,
                        )

                # log for removal/tracking & return
                files_saved[size] = (target_filename, bucket_name)

            if "@archive" in target_filenames:
                size = "@archive"
                (target_filename, _bucket_name) = target_filenames[size]
                # get the calculated bucket_name
                bucket_name = s3_bucketnames[_bucket_name]

                log.debug("Uploading `%s` to `%s` " % (target_filename, bucket_name))

                # calculate the headers ;
                # no need to set acl, its going to be owner-only by default
                _boto3_ExtraArgs = self._boto3_ExtraArgs_default_archive.copy()
                assert resizerResultset.original.format
                _boto3_ExtraArgs["ContentType"] = utils.PIL_type_to_content_type(
                    resizerResultset.original.format
                )

                if not dry_run:
                    # the active file
                    _wrapped = resizerResultset.original

                    # dirty workaround. boto3 has a bug in which it closes files.
                    _buffer = NonCloseableBufferedReader(_wrapped.file)

                    # upload
                    response = self.s3_client.upload_fileobj(  # noqa: F841
                        _buffer,
                        bucket_name,
                        target_filename,
                        ExtraArgs=_boto3_ExtraArgs,
                    )
                    _buffer.detach()

                    # log to external plugin too
                    if self._saverLogger:
                        self._saverLogger.log_save(
                            bucket_name=bucket_name,
                            key=target_filename,
                            file_size=_wrapped.file_size,
                            file_md5=_wrapped.file_md5,
                        )

                # log for removal/tracking & return
                files_saved[size] = (target_filename, bucket_name)

        except Exception as exc:
            # if we have ANY issues, we want to delete everything from amazon s3. otherwise this stuff is just hiding up there
            log.debug(
                "Error uploading... rolling back s3 items. encounted `%s` in `saver.s3.SaverManager.files_save`",
                exc,
            )
            files_saved = self.files_delete(files_saved)
            raise errors.ImageError_S3Upload("error uploading")

        return files_saved


class SaverSimpleAccess(_SaverCoreManager, _core.SaverSimpleAccess):
    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_S3,
    ):
        super(SaverSimpleAccess, self).__init__(
            saverConfig=saverConfig,
            saverLogger=saverLogger,
            resizerConfig=resizerConfig,
        )

    def file_save(
        self,
        bucket_name: str,
        filename: str,
        wrappedFile: BasicImage,
        upload_type: str = "public",
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        if upload_type not in ("public", "archive"):
            raise ValueError("upload_type must be `public` or `archive`")

        if boto3 is None:
            raise NO_BOTO

        # preload the calculated s3 bucketnames
        s3_bucketnames = self.s3_bucketnames

        files_saved = {}
        try:
            # get the calculated bucket_name
            bucket_name = s3_bucketnames[bucket_name]

            log.debug("Uploading `%s` to `%s` " % (filename, bucket_name))

            # calculate the headers ;
            # no need to set acl, its going to be owner-only by default
            _boto3_ExtraArgs = self._boto3_ExtraArgs_default_public.copy()
            assert isinstance(wrappedFile.format, str)
            _boto3_ExtraArgs["ContentType"] = utils.PIL_type_to_content_type(
                wrappedFile.format
            )

            if not dry_run:
                # dirty workaround. boto3 has a bug in which it closes files.
                _buffer = NonCloseableBufferedReader(wrappedFile.file)

                # upload
                response = self.s3_client.upload_fileobj(  # noqa: F841
                    _buffer,
                    bucket_name,
                    filename,
                    ExtraArgs=_boto3_ExtraArgs,
                )
                _buffer.detach()

                # log to external plugin too
                if self._saverLogger:
                    self._saverLogger.log_save(
                        bucket_name=bucket_name,
                        key=filename,
                        file_size=wrappedFile.file_size,
                        file_md5=wrappedFile.file_md5,
                    )

            # log for removal/tracking & return
            files_saved = self.simple_saves_mapping(bucket_name, filename)

            return files_saved
        except Exception as exc:
            log.debug(
                "encountered unexpected exception `%s` in `saver.s3.SaverSimpleAccess.files_save`",
                exc,
            )
            raise

    def simple_saves_mapping(
        self,
        bucket_name: str,
        filename: str,
    ) -> TYPE_files_mapping:
        files_saved: TYPE_files_mapping = {}
        files_saved["`%s`||`%s`" % (bucket_name, filename)] = (filename, bucket_name)
        return files_saved



"""
consolidated way to figure out what files/filetypes are possible

this library once tried to support StringIO.StringIO; bugs with the botocore
and s3transfer library require a workaround that necessitates io
"""
# stdlib
import cgi
import io
from typing import Type
from typing import Union

# ==============================================================================

_CoreFileTypes = (io.IOBase,)
_DefaultMemoryType = io.BytesIO
# _FallbackFileType = tempfile.SpooledTemporaryFile
_FilelikePreference = io.BytesIO
FileReadArgs = "rb"
FileWriteArgs = "wb"

TYPES_FilelikeSupported = Type[io.BytesIO]
TYPES_imagefile_a = Union[
    cgi.FieldStorage,
    io.IOBase,
]



# stdlib
import io
import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# local
from . import errors
from . import image_wrapper
from . import utils


# ==============================================================================

log = logging.getLogger(__name__)

TYPE_ResizesSchema = Dict[str, image_wrapper.ResizerInstructions]
TYPE_selected_resizes = Union[List[str], Tuple[str]]
TYPE_resizes = Dict[str, image_wrapper.BasicImage]

# ------------------------------------------------------------------------------


class ResizerConfig(object):
    """ResizerFactory allows you to specify what/how to resize.

    You could subclass this configuator - just instantiate the object with
        `is_subclass = True`
    to preserve your vars, or configure one on the fly with __init__()

    `resizesSchema` - a dict in this format:
        {  'size_name': {
                'width': 120,
                'height': 120,
                'constraint-method': 'fit-within',
                'save_quality': 50,
                'filename_template': '%(guid)s.%(format)s',
                'suffix': 't1',
                'format':'JPEG',
            },
            'other_size_name': {...},
        }

    `selected_resizes`: an array of size names (see above) to be resized
        if a list or tuple is supplied, it will be copied.

    width*
        in pixels

    height*
        in pixels

    format
        defaults to JPEG

    constraint-method
        see below for valid constraint methods


    save_
        keys prepended with `save_` are stripped of "save_" and are then
        passed on to PIL as kwargs.
        warning: different formats accept different arguments. view the
        code in 'resize' to see what works.

    valid constraint methods:

        see `imagehelper.image_wrapper.ImageWrapper().resize()` for full
        details

        'exact:no-resize'
        'exact:proportion'
        'fit-within'
        'fit-within:crop-to'
        'fit-within:ensure-height'
        'fit-within:ensure-width'
        'smallest:ensure-minimum'

        `optimize` - True / False

    """

    resizesSchema: TYPE_ResizesSchema
    selected_resizes: TYPE_selected_resizes
    optimize_original: Optional[bool] = None
    optimize_resized: bool = False
    # original_allow_animated = None

    def __init__(
        self,
        resizesSchema: Optional[TYPE_ResizesSchema] = None,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        is_subclass: bool = False,
        optimize_original: Optional[bool] = None,
        optimize_resized: bool = False,
        # original_allow_animated=None,
    ):
        if not is_subclass:
            if resizesSchema:
                self.resizesSchema = resizesSchema
            self.optimize_original = optimize_original
            self.optimize_resized = optimize_resized
            # self.original_allow_animated = original_allow_animated

            # we want a unique list
            if selected_resizes is None:
                if resizesSchema is not None:
                    selected_resizes = list(resizesSchema.keys())
            else:
                selected_resizes = selected_resizes[:]
            if selected_resizes:
                self.selected_resizes = list(set(selected_resizes))


class ResizerFactory(object):
    """This is a conveniece Factory to store application configuration
    options.

    You can create a single ResizerFactory in your application, then
    just use it to continually resize images.  Factories have no state, they
    simply hold configuration information, so they're threadsafe.
    """

    resizerConfig: ResizerConfig

    def __init__(self, resizerConfig: ResizerConfig):
        """
        args
            `resizerConfig`
                a resizer.ResizerConfig instance
        """
        self.resizerConfig = resizerConfig

    def resizer(
        self,
        imagefile: Optional[io.IOBase] = None,
        file_b64: Optional[bytes] = None,
    ) -> "Resizer":
        """Returns a resizer object; optionally with an imagefile.
        This does not resize.

        This is useful for validating a file for it's ability to be resized.

        args
            `imagefile`
                an object supported by image_wrapper
                usually:
                    file
                    cgi.fi
            `file_b64`
                b64 encoding of the image file. this is to support serialized
                messagebrokers for workers like celery
        """
        resizer = Resizer(resizerConfig=self.resizerConfig)
        if imagefile is not None and file_b64 is not None:
            raise ValueError("Only pass in `imagefile` or `file_b64`")
        if (imagefile is not None) or (file_b64 is not None):
            resizer.register_image_file(imagefile=imagefile, file_b64=file_b64)
        return resizer


class ResizerResultset(object):
    """A resultset contains two attributes:
    .original  - image_wrapper.BasicImage
    .resizes  - dict.  keys = 'sizes', values = image_wrapper.BasicImage
    """

    resized: TYPE_resizes
    original: image_wrapper.BasicImage

    def __init__(self, resized, original=None):
        self.resized = resized
        self.original = original


class Resizer(object):
    """Resizer is our workhorse.
    It stores the image file, the metadata, and the various resizes."""

    _resizerConfig: Optional[ResizerConfig] = None
    _resizerResultset: Optional[ResizerResultset] = None
    _wrappedImage: Optional[image_wrapper.ImageWrapper]

    def __init__(
        self,
        resizerConfig: Optional[ResizerConfig] = None,
    ):
        self._resizerConfig = resizerConfig
        self._resizerResultset = None
        self._wrappedImage = None

    def register_image_file(
        self,
        imagefile=None,
        imageWrapper: Optional[image_wrapper.ImageWrapper] = None,
        file_b64: Optional[bytes] = None,
        optimize_original: Optional[bool] = None,
    ) -> None:
        """
        registers a file to be resized

        if we pass in cgi.FieldStorage, it seems to bool() to None even when there is a value
        the workaround (grr) is to check against None

        args:
            `imagefile`
                the image as a file
            `imageWrapper`
                the image wrapped in a `imagehelper.image_wrapper.ImageWrapper`
            `file_b64`
                b64 encoding of the image file. this is to support serialized
                messagebrokers for workers like celery
        """
        if self._wrappedImage is not None:
            raise errors.ImageError_DuplicateAction(
                "We already have registered a file."
            )

        if (imagefile is None) and (imageWrapper is None) and (file_b64 is None):
            raise errors.ImageError_ConfigError(
                "Must submit either imagefile /or/ imageWrapper /or/ file_b64"
            )

        if (
            (imagefile is not None)
            and (imageWrapper is not None)
            and (file_b64 is not None)
        ):
            raise errors.ImageError_ConfigError(
                "Submit only imagefile /or/ imageWrapper /or/ file_b64"
            )

        if file_b64 is not None:
            imagefile = utils.b64_decode_to_file(file_b64)

        if imagefile is not None:
            self._wrappedImage = image_wrapper.ImageWrapper(imagefile=imagefile)

        elif imageWrapper is not None:
            if not isinstance(imageWrapper, image_wrapper.ImageWrapper):
                raise errors.ImageError_ConfigError(
                    "imageWrapper must be of type `imaage_wrapper.ImageWrapper`"
                )
            self._wrappedImage = imageWrapper

        if optimize_original is None:
            if self._resizerConfig:
                optimize_original = self._resizerConfig.optimize_original
            else:
                raise ValueError("no optimize_original and no self._resizerConfig")

        if optimize_original:
            # call a standardized interface
            self.optimize_original()

    def optimize_original(self) -> None:
        """standardized inferface for optimizing"""
        log.debug("Resizer.optimize_original")
        assert self._wrappedImage
        self._wrappedImage.basicImage.optimize()

    def resize(
        self,
        imagefile=None,
        imageWrapper: Optional[image_wrapper.ImageWrapper] = None,
        file_b64: Optional[bytes] = None,
        resizesSchema: Optional[TYPE_ResizesSchema] = None,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        optimize_original: Optional[bool] = None,
        optimize_resized: Optional[bool] = None,
        # original_allow_animated=None,
    ) -> ResizerResultset:
        """
        Returns a dict of resized images
        calls self.register_image_file() if needed

        this resizes the images.
        it returns the images and updates the internal dict.

        the internal dict will have an @archive object as well

        args:
            `imagefile`
                the image as a file
            `imageWrapper`
                the image wrapped in a `imagehelper.image_wrapper.ImageWrapper`
            `file_b64`
                b64 encoding of the image file. this is to support serialized
                messagebrokers for workers like celery
        """
        if resizesSchema is None:
            if self._resizerConfig:
                resizesSchema = self._resizerConfig.resizesSchema
            else:
                raise ValueError("no resizesSchema and no self._resizerConfig")
        assert resizesSchema

        if optimize_original is None:
            if self._resizerConfig:
                optimize_original = self._resizerConfig.optimize_original
            else:
                raise ValueError("no optimize_original and no self._resizerConfig")

        if optimize_resized is None:
            if self._resizerConfig:
                optimize_resized = self._resizerConfig.optimize_resized
            else:
                raise ValueError("no optimize_resized and no self._resizerConfig")

        if selected_resizes is None:
            if self._resizerConfig:
                selected_resizes = self._resizerConfig.selected_resizes
            else:
                raise ValueError("no selected_resizes and no self._resizerConfig")
        assert selected_resizes

        if not len(resizesSchema.keys()):
            raise errors.ImageError_ConfigError("We have no resizesSchema...  error")

        if not len(selected_resizes):
            raise errors.ImageError_ConfigError("We have no selected_resizes...  error")

        if (
            (imagefile is not None)
            or (imageWrapper is not None)
            or (file_b64 is not None)
        ):
            self.register_image_file(
                imagefile=imagefile,
                imageWrapper=imageWrapper,
                file_b64=file_b64,
                optimize_original=optimize_original,
            )
        else:
            if optimize_original:
                self.optimize_original()

        if not self._wrappedImage:
            raise errors.ImageError_ConfigError(
                "Please pass in a `imagefile` if you have not set an imageFileObject yet"
            )

        # we'll stash the items here
        resized = {}
        for size in selected_resizes:
            if size[0] == "@":
                raise errors.ImageError_ConfigError(
                    "@ is a reserved initial character for image sizes"
                )

            # ImageWrapper.resize returns a ResizedImage that has attributes `.resized_image`, `image_format`
            resized[size] = self._wrappedImage.resize(resizesSchema[size])
            if optimize_resized:
                resized[size].optimize()

        resizerResultset = ResizerResultset(
            resized=resized, original=self._wrappedImage.get_original()
        )
        self._resizerResultset = resizerResultset

        return resizerResultset

    def fake_resize(
        self,
        original_filename: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
    ) -> ResizerResultset:
        if not self._resizerConfig:
            raise ValueError(
                "fake_resize requires an instance configured with resizerConfig"
            )
        resizesSchema = self._resizerConfig.resizesSchema
        assert resizesSchema

        if selected_resizes is None:
            selected_resizes = self._resizerConfig.selected_resizes
            assert selected_resizes

        if not len(resizesSchema.keys()):
            raise errors.ImageError_ConfigError("We have no resizesSchema...  error")

        if not len(selected_resizes):
            raise errors.ImageError_ConfigError("We have no selected_resizes...  error")

        # we'll stash the items here
        resized = {}
        _original_format = (original_filename.split("."))[-1]
        for size in selected_resizes:
            if size[0] == "@":
                raise errors.ImageError_ConfigError(
                    "@ is a reserved initial character for image sizes"
                )
            _format = resizesSchema[size]["format"]
            _format = utils.derive_format(_format, _original_format)
            resized[size] = image_wrapper.FakedResize(
                format=_format,
                width=resizesSchema[size]["width"],
                height=resizesSchema[size]["height"],
            )

        resizerResultset = ResizerResultset(
            resized=resized,
            original=image_wrapper.FakedOriginal(original_filename=original_filename),
        )
        self._resizerResultset = resizerResultset

        return resizerResultset

    def get_original(self):
        """get the original image, which may have data for us"""
        return self._wrappedImage.get_original()



# stdlib
import cgi
import io
import logging
import tempfile
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

# pypi
try:
    from PIL import Image
except ImportError:
    raise ImportError("Image library (Pillow) is required")
import envoy
from typing_extensions import NotRequired
from typing_extensions import TypedDict

# local
from . import _io
from . import errors
from . import utils

# ==============================================================================

log = logging.getLogger(__name__)


try:
    # PIL.__version__>=9.0.0
    ANTIALIAS = Image.Resampling.LANCZOS
except AttributeError:
    # PIL.__version__<9.0.0
    # py3.6
    ANTIALIAS = Image.ANTIALIAS  # type: ignore [attr-defined]


# import PIL
# _pil_version = PIL.__version__.split(".")
# if int(_pil_version[0]) < 9:
#     # this should only be py36
#     ANTIALIAS = Image.ANTIALIAS
# else:
#     ANTIALIAS = Image.Resampling.LANCZOS


USE_THUMBNAIL: bool = False

_valid_types = [
    cgi.FieldStorage,
    _io._FilelikePreference,
    tempfile.SpooledTemporaryFile,
]
_valid_types.extend(list(_io._CoreFileTypes))
_valid_types_nameless = [_io._FilelikePreference, tempfile.SpooledTemporaryFile]

VALID_TYPES = tuple(_valid_types)
VALID_TYPES_NAMELESS = tuple(_valid_types_nameless)


ResizerInstructions = TypedDict(
    "ResizerInstructions",
    {
        # required
        "width": int,
        "height": int,
        "constraint-method": str,
        "save_quality": int,
        "filename_template": str,
        "suffix": str,
        "format": str,
        # optional below
        "allow_animated": NotRequired[bool],
        # optional - Pillow
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#jpeg-saving
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#png-saving
        "save_optimize": NotRequired[Any],
        "save_progressive": NotRequired[Any],
        "save_transparency": NotRequired[Any],
        "save_bits": NotRequired[Any],
        "save_dictionary": NotRequired[Any],
    },
)

# ------------------------------------------------------------------------------

# these are set to True by imagehelper/__init__.py on import
_OPTIMIZE_SUPPORT_DETECTED: Optional[bool] = None
OPTIMIZE_SUPPORT: Dict = {
    "advpng": {
        "available": None,
        "use": True,
        "options": {"level": 4},  # 4 is max
        "binary": None,
        "*autodetect_args": "--help",
    },
    "gifsicle": {
        "available": None,
        "use": True,
        "options": {"level": 3},
        "binary": None,
        "*autodetect_args": "--help",
    },
    "jpegtran": {
        "available": None,
        "use": True,
        "options": {"progressive": True},
        "binary": None,
        "*autodetect_args": "--help",
    },
    "jpegoptim": {
        "available": None,
        "use": True,
        "binary": None,
        "*autodetect_args": "--help",
    },
    "optipng": {
        "available": None,
        "use": True,
        "options": {"level": 3},  # 6 would be best
        "binary": None,
        "*autodetect_args": "--help",
    },
    "pngcrush": {
        "available": None,
        "use": True,
        "binary": None,
        "*autodetect_args": "--help",
    },
}


def autodetect_support(
    test_libraries: Optional[List[str]] = None,
) -> None:
    """
    ``test_libraries`` should be a list of keys in the OPTIMIZE_SUPPORT list
    if not supplied, all will be tested
    this only needs to be done once per application

    it tests if a program exists by invoking the `help` subcommand for
    each shell command. this runs relatively quickly, but calls multiple
    external programs. Python3 has `shutil.which` but this library supports
    Python2

    if you have a forking application, you should run before forking:

        import imagehelper
        imagehelper.image_wrapper.autodetect_support()

    otherwise, the first invocation of .optimize() will cause this to run
    once run, `_OPTIMIZE_SUPPORT_DETECTED` will be set, blocking further
    autodetections
    """
    log.debug("autodetect_support")
    if test_libraries is None:
        test_libraries = sorted(OPTIMIZE_SUPPORT.keys())
    for library in test_libraries:
        _binary = OPTIMIZE_SUPPORT[library]["binary"] or library
        result = envoy.run(
            """%s %s""" % (_binary, OPTIMIZE_SUPPORT[library]["*autodetect_args"])
        )
        if result.status_code == 1:
            log.debug("   available:    %s", library)
            OPTIMIZE_SUPPORT[library]["available"] = True
        elif result.status_code == 127:
            log.debug("!!! unavailable: %s", library)
            OPTIMIZE_SUPPORT[library]["available"] = False
        elif result.status_code == 0:
            # on ubuntu:
            # gifsicle  - stdout
            # optipng   - stdout
            # jpegoptim - stderr
            txt = result.std_out.lower() or result.std_err.lower()
            if ("--" in txt) and (library in txt):
                log.debug("   available:    %s", library)
                OPTIMIZE_SUPPORT[library]["available"] = True
            else:
                log.debug("xxx unsupported code: %s : %s", library, result.status_code)
        else:
            log.debug("xxx unsupported code: %s : %s", library, result.status_code)
            # leave None for every other code
        OPTIMIZE_SUPPORT[library]["*autodetect-status_code"] = result.status_code
    global _OPTIMIZE_SUPPORT_DETECTED
    _OPTIMIZE_SUPPORT_DETECTED = True


# ==============================================================================


class BasicImage(object):
    """
    A generic wrapper for Images
    """

    # `file` a file-like object; ie, StringIO
    # file: io.IOBase
    file: io.BytesIO

    # `format`
    format: Optional[str] = None

    # `name`
    name: Optional[str] = None

    # `mode` file attribute
    mode = None

    # `width` file attribute
    width: Optional[int] = None

    # `height` file attribute
    height: Optional[int] = None

    # `None` by default; `True` if `optimize` successful; `False` if it failed
    is_optimized: Optional[bool] = None

    # `None` by default. If `optimize` is run, it becomes a list of external tool + status
    optimizations: Optional[List] = None

    def __init__(
        self,
        fileObject: io.BytesIO,
        name: Optional[str] = None,
        format: Optional[str] = None,
        mode=None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        is_image_animated: Optional[bool] = None,
        animated_image_totalframes: Optional[int] = None,
    ):
        """
        :arg fileObject: the resized file
        :param name: default None
        :param format: default None
        :param mode: default None
        :param width: default None
        :param height: default None
        :param is_image_animated: default None
        :param animated_image_totalframes: default None
        """
        self.file = fileObject
        self.file.seek(0)  # be kind, rewind
        self.name = name
        self.format = format
        self.mode = mode
        self.width = width
        self.height = height
        self.optimization_savings = 0
        self.is_image_animated = is_image_animated
        self.animated_image_totalframes = animated_image_totalframes

    @property
    def file_size(self) -> int:
        """property; calculate the file's size in bytes"""
        return utils.file_size(self.file)

    @property
    def file_md5(self) -> str:
        """property; calculate the file's md5"""
        return utils.file_md5(self.file)

    @property
    def file_b64(self) -> bytes:
        """property; base64 encode the file"""
        return utils.file_b64(self.file)

    @property
    def format_standardized(self) -> Optional[str]:
        """proxied format; standardized version"""
        if self.format is None:
            return None
        return utils.PIL_type_to_standardized(self.format)

    @property
    def file_extension(self) -> Optional[str]:
        """proxied format; PIL version"""
        if self.format is None:
            return None
        return utils.PIL_type_to_extension(self.format)

    def optimize(
        self,
    ) -> None:
        """this does some heavy lifting

        unix/mac only feature; sorry.

        this function creates an infile and outfile via NamedTemporaryFile
        it then pipes the file through lossless compression options

        this will replace the self.file object
        """
        if self.format_standardized not in ("jpg", "png", "gif"):
            return
        log.debug("optimizing a file.  format is: %s" % self.format_standardized)

        FilelikePreference = _io._FilelikePreference

        # we need to write the image onto the disk with an infile and outfile
        # this does suck.
        self.file.seek(0)
        fileInput = tempfile.NamedTemporaryFile()
        if hasattr(self.file, "getvalue"):
            fileInput.write(self.file.getvalue())
        elif hasattr(self.file, "read"):
            fileInput.write(self.file.read())
        else:
            raise ValueError(
                "not sure what to do; this `file` object is not what I expected."
            )
        fileInput.seek(0)
        fileOutput = tempfile.NamedTemporaryFile()  # keep this open for the next block

        _fname_input = fileInput.name
        _fname_output = fileOutput.name

        # we need this for filesavings
        filesize_original = utils.file_size(fileInput)

        # run the autodetect
        if not _OPTIMIZE_SUPPORT_DETECTED:
            autodetect_support()

        _optimized = False
        optimizations = []
        if self.format_standardized == "jpg":
            if (
                OPTIMIZE_SUPPORT["jpegtran"]["available"]
                and OPTIMIZE_SUPPORT["jpegtran"]["use"]
            ):
                _binary = OPTIMIZE_SUPPORT["jpegtran"]["binary"] or "jpegtran"
                _progressive = (
                    "-progressive"
                    if OPTIMIZE_SUPPORT["jpegtran"]["options"]["progressive"]
                    else ""
                )
                r = envoy.run(
                    """%s -copy all -optimize %s -outfile %s %s"""
                    % (_binary, _progressive, _fname_output, _fname_input)
                )
                if r.status_code != 127:
                    _optimized = True
                    optimizations.append(("jpegtran", True))
                else:
                    optimizations.append(("jpegtran", False))

            if (
                OPTIMIZE_SUPPORT["jpegoptim"]["available"]
                and OPTIMIZE_SUPPORT["jpegoptim"]["use"]
            ):
                print("optimizing with jpegoptim")
                _binary = OPTIMIZE_SUPPORT["jpegoptim"]["binary"] or "jpegoptim"
                r = envoy.run("""%s --strip-all -q %s""" % (_binary, _fname_output))
                if r.status_code != 127:
                    _optimized = True
                    optimizations.append(("jpegoptim", True))
                else:
                    optimizations.append(("jpegoptim", False))

        elif self.format_standardized == "gif":
            if (
                OPTIMIZE_SUPPORT["gifsicle"]["available"]
                and OPTIMIZE_SUPPORT["gifsicle"]["use"]
            ):
                _binary = OPTIMIZE_SUPPORT["gifsicle"]["binary"] or "gifsicle"
                _gifsicle_level = OPTIMIZE_SUPPORT["gifsicle"]["options"]["level"]
                r = envoy.run(
                    """%s -O%d %s --output %s"""
                    % (_binary, _gifsicle_level, _fname_input, _fname_output)
                )
                if r.status_code != 127:
                    _optimized = True
                    optimizations.append(("gifsicle", True))
                else:
                    optimizations.append(("gifsicle", False))

        elif self.format_standardized == "png":
            if (
                OPTIMIZE_SUPPORT["pngcrush"]["available"]
                and OPTIMIZE_SUPPORT["pngcrush"]["use"]
            ):
                _binary = OPTIMIZE_SUPPORT["pngcrush"]["binary"] or "pngcrush"
                # envoy.run("""pngcrush -rem alla -reduce -brute -q %s %s""" % (_fname_input, _fname_output))
                # envoy.run("""pngcrush -rem alla -reduce -q %s %s""" % (_fname_input, _fname_output))
                r = envoy.run(
                    """%s -rem alla -nofilecheck -bail -blacken -reduce -cc %s %s"""
                    % (_binary, _fname_input, _fname_output)
                )
                if r.status_code != 127:
                    _fname_input = _fname_output
                    _optimized = True
                    optimizations.append(("pngcrush", True))
                else:
                    optimizations.append(("pngcrush", False))

            if (
                OPTIMIZE_SUPPORT["optipng"]["available"]
                and OPTIMIZE_SUPPORT["optipng"]["use"]
            ):
                _binary = OPTIMIZE_SUPPORT["optipng"]["binary"] or "optipng"
                _optipng_level = OPTIMIZE_SUPPORT["optipng"]["options"]["level"]
                # note that we do `--out OUTPUT --(stop) INPUT
                r = envoy.run(
                    """%s -i0 -o%d -out %s -- %s"""
                    % (_binary, _optipng_level, _fname_output, _fname_input)
                )
                if r.status_code != 127:
                    _fname_input = _fname_output
                    _optimized = True
                    optimizations.append(("optipng", True))
                else:
                    optimizations.append(("optipng", False))

            if (
                OPTIMIZE_SUPPORT["advpng"]["available"]
                and OPTIMIZE_SUPPORT["advpng"]["use"]
            ):
                _binary = OPTIMIZE_SUPPORT["advpng"]["binary"] or "advpng"
                _advpng_level = OPTIMIZE_SUPPORT["advpng"]["options"]["level"]
                # note that we do `--out OUTPUT --(stop) INPUT
                r = envoy.run(
                    """%s -%d -z %s""" % (_binary, _advpng_level, _fname_output)
                )
                if r.status_code != 127:
                    _fname_input = _fname_output
                    _optimized = True
                    optimizations.append(("advpng", True))
                else:
                    optimizations.append(("advpng", False))

        # stash this onto the object
        self.optimizations = optimizations

        if not _optimized:
            self.file.seek(0)
            self.is_optimized = False  # nothing to optimize
            log.debug("optimization_savings = 0 - NO OPTIMIZATIONS POSSIBLE")
            self.optimization_savings = 0

        else:
            fileOutput.seek(0)
            newFile = FilelikePreference()
            newFile.write(fileOutput.read())
            newFile.seek(0)
            self.file = newFile
            self.is_optimized = True

            # so how much did we save?
            filesize_optimized = utils.file_size(newFile)
            optimization_savings = filesize_original - filesize_optimized

            log.debug("optimization_savings = %s" % optimization_savings)
            self.optimization_savings = optimization_savings

        # done with these, so close
        fileInput.close()
        fileOutput.close()


class ResizedImage(BasicImage):
    """A class for a ResizedImage Result."""

    def __repr__(self):
        return "<ReizedImage at %s - %s >" % (id(self), self.__dict__)


class FakedOriginal(BasicImage):
    """sometimes we need to fake an original file"""

    format: str
    mode = None
    width: int
    height: int
    file_size: int
    file_md5: str

    def __init__(self, original_filename: str):
        file_ext = original_filename.split(".")[-1].lower()
        self.format = utils.standardized_to_PIL_type(file_ext)


class FakedResize(BasicImage):
    file: None = None  # type: ignore [assignment]
    format: str
    width: int
    height: int

    def __init__(
        self,
        format: str,
        width: int,
        height: int,
    ):
        self.format = format
        self.width = width
        self.height = height


class ImageWrapper(object):
    """Our base class for image operations"""

    basicImage: BasicImage
    pilObject: Image.Image

    def get_original(self):
        return self.basicImage

    def __del__(self):
        if self.pilObject is not None:
            self.pilObject.close()

    def __init__(
        self,
        imagefile: io.IOBase,
        imagefile_name: Optional[str] = None,
        FilelikePreference: Optional[_io.TYPES_FilelikeSupported] = None,
    ):
        """
        registers and validates the image file
        note that we do copy the image file

        args:

        `imagefile`
                cgi.FieldStorage
                _io._CoreFileTypes = [io.IOBase, ]
                _io._FilelikePreference
                tempfile.TemporaryFile, tempfile.SpooledTemporaryFile

        `imagefile_name`
            only used for informational purposes

        `FilelikePreference`
            preference class for filelike objects
                _io._FilelikePreference
                tempfile.SpooledTemporaryFile
        """
        if imagefile is None:
            raise errors.ImageError_MissingFile(utils.ImageErrorCodes.MISSING_FILE)

        if not isinstance(imagefile, VALID_TYPES):
            raise errors.ImageError_Parsing(
                utils.ImageErrorCodes.UNSUPPORTED_IMAGE_CLASS
            )

        try:
            # try to cache this all
            file_data = None
            file_name = None
            if isinstance(imagefile, cgi.FieldStorage):
                if not hasattr(imagefile, "filename"):
                    raise errors.ImageError_Parsing(
                        utils.ImageErrorCodes.MISSING_FILENAME_METHOD
                    )
                imagefile.file.seek(0)
                file_data = imagefile.file.read()

                file_name = ""
                if hasattr(imagefile.file, "name"):
                    file_name = imagefile.file.name
                elif hasattr(imagefile, "filename"):
                    file_name = imagefile.filename

                # be kind, rewind; the input obj we no longer care about
                # but someone else might care
                imagefile.file.seek(0)

            elif isinstance(imagefile, VALID_TYPES_NAMELESS):
                imagefile.seek(0)
                file_data = imagefile.read()
                file_name = imagefile_name or ""

                # be kind, rewind; the input obj we no longer care about
                # but someone else might care
                imagefile.seek(0)

            elif isinstance(imagefile, _io._CoreFileTypes):
                # catch this last
                imagefile.seek(0)
                file_data = imagefile.read()
                # default
                file_name = imagefile_name
                if hasattr(imagefile, "name"):  # Py3 object may not have this
                    file_name = imagefile.name
                if file_name == "<fdopen>":
                    file_name = imagefile_name or ""

                # be kind, rewind; the input obj we no longer care about
                # but someone else might care
                imagefile.seek(0)

            else:
                # just be safe with an extra else
                log.debug(
                    "isinstance(imagefile) is not expected. type() is `%s`",
                    type(imagefile),
                )
                raise errors.ImageError_Parsing(
                    utils.ImageErrorCodes.UNSUPPORTED_IMAGE_CLASS
                )

            if FilelikePreference is None:
                FilelikePreference = _io._FilelikePreference

            # create a new image
            # and stash our data!
            fh_imageData = FilelikePreference()
            fh_imageData.write(file_data)
            fh_imageData.seek(0)
            fh_name = imagefile_name or file_name

            # make the new wrapped obj and then...
            # safety first! just ensure this loads.
            pilObject = Image.open(fh_imageData)
            pilObject.load()
            if not pilObject:
                raise errors.ImageError_Parsing(utils.ImageErrorCodes.INVALID_REBUILD)
            self.pilObject = pilObject

            # finally, stash our data
            wrappedImage = BasicImage(
                fh_imageData,
                name=fh_name,
                format=self.pilObject.format,
                mode=self.pilObject.mode,
                width=self.pilObject.size[0],
                height=self.pilObject.size[1],
                is_image_animated=utils.is_image_animated(self.pilObject),
                animated_image_totalframes=utils.animated_image_totalframes(
                    self.pilObject
                ),
            )
            self.basicImage = wrappedImage

        except IOError as exc:
            log.debug("encountered an IOError. Exception is: `%s`", exc)
            raise errors.ImageError_Parsing(utils.ImageErrorCodes.INVALID_FILETYPE)

        except errors.ImageError as exc:
            log.debug("encountered ImageError exception: `%s`", exc)
            raise

        except Exception as exc:
            log.debug("encountered unknown exception: `%s`", exc)
            raise

    def resize(
        self,
        instructions_dict: ResizerInstructions,
        FilelikePreference: Optional[_io.TYPES_FilelikeSupported] = None,
    ) -> ResizedImage:
        """this does the heavy lifting

        be warned - this uses a bit of memory!

        1. we operate on a copy of the pilObject via _io.cStringIo
            (which is already a copy of the original)
        2. we save to another new _io 'file'

        valid `constraint-method` for `instructions_dict`

            'fit-within'
                Resizes item to fit within the bounding box, on both height
                and width.   This resulting image will be the size of the
                bounding box or smaller.

            'fit-within:crop-to'
                resizes the item along whichever axis ensures the bounding box
                is 100% full, then crops.  This resulting image will be the
                size of the bounding box.

            'fit-within:ensure-width'
                resizes item to fit within the bounding box, scaling height
                to ensure 100% width.  This resulting image will be the size of
                the bounding box.

            'fit-within:ensure-height'
                resizes item to fit within the bounding box, scaling width to
                ensure 100% height. This resulting image will be the size of
                the bounding box.

            'smallest:ensure-minimum'
                useful for things like og:image where you want at least a 200px
                image.

            'exact:no-resize'
                don't scale! raises an error if a scale must be made. this is a
                convenience for just saving/re-encoding files.
                i.e. 100x100 must receive an image that is 100x100

            'exact:proportion'
                tries to scale the image to an exact size.  raises an error if
                it can't.  Usually this is used to resample a 1:1 image, however
                this might be used to drop an image to a specific proportion.
                i.e. 300x400 can scale to 30x40, 300x400 but not 30x50

            'passthrough:no-resize'
                don't scale!

            `FilelikePreference` - default preference for file-like objects
        """
        # mypy typing
        if self.pilObject is None:
            raise ValueError("mising `self.pilObject`")

        if FilelikePreference is None:
            FilelikePreference = _io._FilelikePreference

        # we analyze the pilObject, because `copy()` only works on the frame
        if utils.is_image_animated(self.pilObject):
            allow_animated = False
            if "allow_animated" in instructions_dict:
                allow_animated = instructions_dict["allow_animated"]
            if not allow_animated:
                raise errors.ImageError_InstructionsError(
                    "Image is Animated but instructions do not allow it!"
                )

        resized_image = self.pilObject.copy()

        if resized_image.palette:
            resized_image = resized_image.convert()

        constraint_method = "fit-within"
        if "constraint-method" in instructions_dict:
            constraint_method = instructions_dict["constraint-method"]

        if constraint_method != "passthrough:no-resize":
            # t_ = target
            # i_ = image / real

            (i_w, i_h) = self.pilObject.size

            t_w = instructions_dict["width"]
            t_h = instructions_dict["height"]

            crop: Optional[Tuple[int, int, int, int]] = None
            # type hints for below switch
            proportion: float
            proportion_w: float
            proportion_h: float

            # notice that we only scale DOWN (ie: check that t_x < i_x

            if constraint_method in ("fit-within", "fit-within:crop-to"):
                # figure out the proportions
                proportion_w = 1
                proportion_h = 1
                if t_w < i_w:
                    proportion_w = t_w / i_w
                if t_h < i_h:
                    proportion_h = t_h / i_h

                if constraint_method == "fit-within":
                    # peg to the SMALLEST proportion so the entire image fits
                    if proportion_w < proportion_h:
                        proportion_h = proportion_w
                    elif proportion_h < proportion_w:
                        proportion_w = proportion_h
                    # figure out the resizes!
                    t_w = int(i_w * proportion_w)
                    t_h = int(i_h * proportion_h)

                elif constraint_method == "fit-within:crop-to":
                    # peg so the smallest dimension fills the canvas, then crop the rest.
                    if proportion_w > proportion_h:
                        proportion_h = proportion_w
                    elif proportion_h > proportion_w:
                        proportion_w = proportion_h

                    # note what we want to crop to
                    crop_w = t_w
                    crop_h = t_h

                    # figure out the resizes!
                    t_w = int(i_w * proportion_w)
                    t_h = int(i_h * proportion_h)

                    if (crop_w != t_w) or (crop_h != t_h):
                        # support_hack_against_artifacting handles an issue where .thumbnail makes stuff look like shit
                        # except we're not using .thumbnail anymore; we're using resize directly
                        support_hack_against_artifacting = USE_THUMBNAIL
                        if support_hack_against_artifacting:
                            if t_w < i_w:
                                t_w += 1
                            if t_h < i_h:
                                t_h += 1

                        (x0, y0, x1, y1) = (0, 0, t_w, t_h)

                        if t_w > crop_w:
                            x0 = int((t_w / 2) - (crop_w / 2))
                            x1 = x0 + crop_w

                        if t_h > crop_h:
                            y0 = int((t_h / 2) - (crop_h / 2))
                            y1 = y0 + crop_h

                        crop = (x0, y0, x1, y1)

            elif constraint_method == "fit-within:ensure-width":
                proportion = 1
                if t_w < i_w:
                    proportion = t_w / i_w
                t_h = int(i_h * proportion)

            elif constraint_method == "fit-within:ensure-height":
                proportion = 1
                if t_h < i_h:
                    proportion = t_h / i_h
                t_w = int(i_w * proportion)

            elif constraint_method == "smallest:ensure-minimum":
                # useful for things like og:image where you want at least a 200px image

                # figure out the proportions
                proportion_w = t_w / i_w
                proportion_h = t_h / i_h

                # we don't want to scale up...
                if proportion_h > 1 or proportion_w > 1:
                    proportion_h = 1
                    proportion_w = 1

                # use_scale = "h"
                scale_factor = proportion_h
                if proportion_w > proportion_h:
                    # use_scale = "w"
                    scale_factor = proportion_w

                t_h = int(i_h * scale_factor)
                t_w = int(i_w * scale_factor)

            elif constraint_method == "exact:proportion":
                proportion_w = 1
                proportion_h = 1
                if t_w < i_w:
                    proportion_w = t_w / i_w
                if t_h < i_h:
                    proportion_h = t_h / i_h
                if proportion_w != proportion_h:
                    raise errors.ImageError_ResizeError(
                        "item can not be scaled to exact size"
                    )

            elif constraint_method == "exact:no-resize":
                if (t_w != i_w) or (t_h != i_h):
                    raise errors.ImageError_ResizeError("item is not exact size")

            else:
                raise errors.ImageError_ResizeError(
                    'Invalid constraint-method for size recipe: "%s"'
                    % constraint_method
                )

            if (i_w != t_w) or (i_h != t_h):
                if USE_THUMBNAIL:
                    # the thumbnail is faster, but has been looking uglier in recent versions
                    resized_image.thumbnail((t_w, t_h), ANTIALIAS)
                else:
                    resized_image = resized_image.resize((t_w, t_h), ANTIALIAS)

            if crop:
                resized_image = resized_image.crop(crop)
                resized_image.load()

        format = "JPEG"
        if "format" in instructions_dict:
            format = instructions_dict["format"]

        # returns uppercase
        format = utils.derive_format(format, self.get_original().format)

        def _get_pil_options(_format: str) -> Dict:
            pil_options: Dict = {}
            if format in ("JPEG", "PDF"):
                for i in ("quality", "optimize", "progressive"):
                    k = "save_%s" % i
                    if k in instructions_dict:
                        pil_options[i] = instructions_dict[k]  # type: ignore[literal-required]
            elif format == "PNG":
                for i in ("optimize", "transparency", "bits", "dictionary"):
                    k = "save_%s" % i
                    if k in instructions_dict:
                        pil_options[i] = instructions_dict[k]  # type: ignore[literal-required]
            return pil_options

        # inner function will generate keys for us
        pil_options = _get_pil_options(format)

        # save the image !
        resized_image_file = FilelikePreference()
        resized_image.save(resized_image_file, format, **pil_options)

        return ResizedImage(
            resized_image_file,
            format=format,
            width=resized_image.size[0],
            height=resized_image.size[1],
        )



import os
import re
from setuptools import find_packages
from setuptools import setup

HERE = os.path.abspath(os.path.dirname(__file__))

# store version in the init.py
with open(os.path.join(HERE, "src", "imagehelper", "__init__.py")) as v_file:
    VERSION = (
        re.compile(r".*__VERSION__ = \"(.*?)\"", re.S).match(v_file.read()).group(1)
    )

description = "Simple utilites for image resizing and uploading and stuff like that."
long_description = """The `imagehelper` package offers a simple interface for image resizing, optimizing and uploading. Core image resizing operations are handled by the `Pillow` (PIL) package; S3 uploading is handled by `boto3`, and there are hooks for optimizing the images with the commandline tools: `advpng`,  `gifsicle`, `jpegtran`, `jpegoptim`, `optipng` and `pngcrush.`"""
with open(os.path.join(HERE, "README.md")) as fp:
    long_description = fp.read()

requires = [
    "envoy",
    "Pillow",
    "typing_extensions",
]
tests_require = [
    "boto3",
    # "botocore",  # part of boto3; listed to upgrade better
    # "certifi",
    "pytest",
    "requests",
    # "urllib3",  # used by botocore; listed to upgrade better
]
testing_extras = tests_require + []

setup(
    name="imagehelper",
    version=VERSION,
    url="http://github.com/jvanasco/imagehelper",
    author="Jonathan Vanasco",
    author_email="jonathan@findmeon.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Graphics :: Graphics Conversion",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    py_modules=["imagehelper"],
    license="BSD",
    packages=find_packages(
        where="src",
    ),
    package_dir={"": "src"},
    package_data={"imagehelper": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    tests_require=tests_require,
    extras_require={
        "testing": testing_extras,
    },
    test_suite="tests",
)



class ImageError(Exception):
    """Base class for Exceptions"""

    pass


class ImageError_ArgsError(ImageError):
    pass


class ImageError_ConfigError(ImageError):
    pass


class ImageError_DuplicateAction(ImageError):
    pass


class ImageError_InstructionsError(ImageError):
    pass


class ImageError_MissingFile(ImageError):
    pass


class ImageError_Parsing(ImageError):
    pass


class ImageError_ResizeError(ImageError):
    pass


class ImageError_SaverUpload(ImageError):
    pass


class ImageError_S3Upload(ImageError_SaverUpload):
    pass



# stdlib
import base64
import hashlib
import logging
import os
from typing import Dict
from typing import Tuple

# pypi
from PIL import ImageSequence

# local
from . import _io

# ==============================================================================

log = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

TYPE_files_mapping = Dict[str, Tuple[str, str]]


class ImageErrorCodes(object):
    """Consolidating codes and error messages"""

    INVALID_FILETYPE = 1
    INVALID_OTHER = 2
    NO_IMAGE = 3
    MISSING_FILE = 4
    UNSUPPORTED_IMAGE_CLASS = 5  # Must be cgi.FieldStorage or file
    INVALID_REBUILD = 6
    MISSING_FILENAME_METHOD = 7


_PIL_type_to_content_type: Dict[str, str] = {
    "gif": "image/gif",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "pdf": "application/pdf",
    "png": "image/png",
}

_PIL_type_to_standardized: Dict[str, str] = {
    "gif": "gif",
    "jpg": "jpg",
    "jpeg": "jpg",
    "pdf": "pdf",
    "png": "png",
}

_standardized_to_PIL_type: Dict[str, str] = {
    "gif": "GIF",
    "jpg": "JPEG",
    "jpeg": "JPEG",
    "pdf": "PDF",
    "png": "PNG",
}


# ==============================================================================


def is_image_animated(im) -> bool:
    im.seek(0)
    try:
        im.seek(1)
        return True
    except EOFError:
        return False


def animated_image_totalframes(im) -> int:
    im.seek(0)
    _frame = 0
    for frame in ImageSequence.Iterator(im):
        _frame = _frame + 1
    return _frame


def derive_format(intended_format: str, original_format: str) -> str:
    """
    1. returns uppercase
    2. for the special cases of "AUTO" and "ORIGINAL"
       this will parse the original format and possibly change it

    SEE ALSO: imagehelper.saver.utils.derive_resized_format
    """
    intended_format = intended_format.upper()
    if intended_format in ("AUTO", "ORIGINAL"):
        _og_format = original_format.upper()
        if _og_format in ("PNG", "GIF"):
            intended_format = "PNG"
        else:
            intended_format = "JPEG"
    return intended_format


def PIL_type_to_content_type(ctype: str) -> str:
    ctype = ctype.lower()
    if ctype in _PIL_type_to_content_type:
        return _PIL_type_to_content_type[ctype]
    raise ValueError("invalid ctype")


def PIL_type_to_standardized(ctype: str) -> str:
    ctype = ctype.lower()
    if ctype in _PIL_type_to_standardized:
        return _PIL_type_to_standardized[ctype]
    raise ValueError("invalid ctype - `%s`" % ctype)


def PIL_type_to_extension(ctype: str) -> str:
    ctype = ctype.lower()
    if ctype in _PIL_type_to_standardized:
        return _PIL_type_to_standardized[ctype]
    raise ValueError("invalid ctype")


def standardized_to_PIL_type(ctype: str) -> str:
    ctype = ctype.lower()
    if ctype in _standardized_to_PIL_type:
        return _standardized_to_PIL_type[ctype]
    raise ValueError("invalid ctype")


def file_size(fileobj) -> int:
    """what's the size of the object?"""
    fileobj.seek(0, os.SEEK_END)
    sized = fileobj.tell()
    fileobj.seek(0)
    return sized


def file_md5(fileobj) -> str:
    fileobj.seek(0)
    md5 = hashlib.md5()
    block_size = md5.block_size * 128
    for chunk in iter(lambda: fileobj.read(block_size), b""):
        md5.update(chunk)
    fileobj.seek(0)
    return md5.hexdigest()


def file_b64(fileobj) -> bytes:
    fileobj.seek(0)
    as_b64 = base64.encodebytes(fileobj.read())
    fileobj.seek(0)
    return as_b64


def b64_decode_to_file(coded_string):
    decoded_data = base64.b64decode(coded_string)
    fileobj = _io._FilelikePreference()
    fileobj.write(decoded_data)
    fileobj.seek(0)
    return fileobj



# stdlib
import logging
import os
from typing import Dict
from typing import Optional

# local
from . import _core
from .utils import check_archive_original
from .utils import size_to_filename
from .. import _io
from .. import errors
from .. import utils
from ..image_wrapper import BasicImage
from ..image_wrapper import ResizerInstructions
from ..resizer import ResizerConfig
from ..resizer import ResizerResultset
from ..resizer import TYPE_selected_resizes
from ..utils import TYPE_files_mapping


# ==============================================================================

log = logging.getLogger(__name__)


class ResizerInstructions_Localfile(ResizerInstructions):
    subdir_public: str


TYPE_ResizesSchema_Localfile = Dict[str, ResizerInstructions_Localfile]


class ResizerConfig_Localfile(ResizerConfig):
    resizesSchema: TYPE_ResizesSchema_Localfile  # type: ignore[assignment]


# ------------------------------------------------------------------------------


class SaverConfig(_core.SaverConfig):
    """
    Configuration for a localfile saving implementation

    we will save to a `filedir`
    """

    subdir_public_name: str
    subdir_archive_name: str
    filedir: str
    archive_original: Optional[bool] = None

    def __init__(
        self,
        subdir_public_name: str = "public",
        subdir_archive_name: str = "archive",
        filedir: str = "localfile-output",
        archive_original: Optional[bool] = None,
    ):
        self.subdir_public_name = subdir_public_name
        self.subdir_archive_name = subdir_archive_name
        self.filedir = filedir
        self.archive_original = archive_original


class SaverLogger(_core.SaverLogger):
    """
    Abstract interface for logging saves.
    This should be subclassed and provided to a factory
    """

    def log_save(  # type: ignore[override]
        self,
        subdir_name: Optional[str] = None,
        filename: Optional[str] = None,
        file_size: Optional[int] = None,
        file_md5: Optional[str] = None,
        filepath: Optional[str] = None,
    ) -> None:
        """
        kwargs:
            subdir_name : name of the subdirectory
            filename : name of the file
            file_size : size of the file in bytes
            file_md5 : md5 value of the file (hex)
            filepath : full filepath for the save. this may be relative but must include the directory above `subdir`
        """
        pass

    def log_delete(  # type: ignore[override]
        self,
        subdir_name: Optional[str] = None,
        filename: Optional[str] = None,
        filepath: Optional[str] = None,
    ) -> None:
        """
        kwargs:
            subdir_name : name of the subdirectory
            filename : name of the file
            filepath : full filepath for the save. this may be relative but must include the directory above `subdir`
        """
        pass


class SaverManagerFactory(_core.SaverManagerFactory):
    """Factory for generating SaverManager instances"""

    _resizerConfig: ResizerConfig_Localfile
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_Localfile,
    ):
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig

    def manager(self) -> "SaverManager":
        """generate and return a new SaverManager instance"""
        return SaverManager(
            saverConfig=self._saverConfig,
            saverLogger=self._saverLogger,
            resizerConfig=self._resizerConfig,
        )

    def simple_access(self) -> "SaverSimpleAccess":
        """generate and return a new SaverSimpleAccess instance"""
        return SaverSimpleAccess(
            saverConfig=self._saverConfig,
            saverLogger=self._saverLogger,
            resizerConfig=self._resizerConfig,
        )


class _SaverCoreManager(_core._SaverCoreManager):
    """
    based on interface defined in `_core.SaverCoreManager`
    """

    _resizerConfig: ResizerConfig_Localfile
    _saverConfig: SaverConfig
    _saverLogger: SaverLogger
    filename_template: str = "%(guid)s-%(suffix)s.%(format)s"
    filename_template_archive: str = "%(guid)s.%(format)s"

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_Localfile,
    ):
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig

    def files_delete(
        self,
        files_saved: TYPE_files_mapping,
        dry_run: bool = False,
        remove_empty_dirs: bool = True,
    ) -> TYPE_files_mapping:
        """does not actually delete"""
        for size in list(files_saved.keys()):
            # grab the stash
            (target_filename, subdir_name) = files_saved[size]

            target_dirname = os.path.join(self._saverConfig.filedir, subdir_name)
            if not os.path.exists(target_dirname):
                raise ValueError("Directory %s does not exist!" % target_dirname)
            target_filepath = os.path.join(target_dirname, target_filename)

            # delete it
            log.debug("going to delete %s " % (target_filepath,))

            if not dry_run:
                os.unlink(target_filepath)

                # external logging
                if self._saverLogger:
                    self._saverLogger.log_delete(
                        subdir_name=subdir_name,
                        filename=target_filename,
                        filepath=target_filepath,
                    )

                # cleanup path
                if remove_empty_dirs:
                    if len(os.listdir(target_dirname)) == 0:
                        os.rmdir(target_dirname)

            # internal cleanup
            del files_saved[size]

        return files_saved


class SaverManager(_SaverCoreManager, _core.SaverManager):
    """
    `SaverManager` handles all the actual uploading and deleting

    based on interface defined in `_core._SaverCoreManager`
    but inherits from this file's '`_SaverCoreManager`
    """

    def __init__(
        self,
        saverConfig: SaverConfig,
        saverLogger: SaverLogger,
        resizerConfig: ResizerConfig_Localfile,
    ):
        if not resizerConfig:
            raise ValueError(
                """`SaverManager` requires a `resizerConfig` which contains the resize recipes. these are needed for generating filenames."""
            )
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig

    def _validate__selected_resizes(
        self,
        resizerResultset: ResizerResultset,
        selected_resizes: Optional[TYPE_selected_resizes],
    ) -> TYPE_selected_resizes:
        """
        shared validation
        returns: `LIST` selected_resizes
        """

        # default to the resized images
        if selected_resizes is None:
            selected_resizes = list(resizerResultset.resized.keys())

        for k in selected_resizes:
            if k not in resizerResultset.resized:
                raise errors.ImageError_ConfigError(
                    "selected size is not resizerResultset.resized (%s)" % k
                )

            if k not in self._resizerConfig.resizesSchema:
                raise errors.ImageError_ConfigError(
                    "selected size is not self._resizerConfig.resizesSchema (%s)" % k
                )

            # exist early for invalid sizes
            if k[0] == "@":
                raise errors.ImageError_ConfigError(
                    "@ is a reserved initial character for image sizes (%s)" % k
                )

        return selected_resizes

    def generate_filenames(
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
    ) -> TYPE_files_mapping:
        """
        generates the filenames s3 would save to;
        this is useful for planning/testing or deleting old files

        Returns a `dict` of target filenames:
            keys = resized size
            values = tuple (target_filename, subdir)
        """
        if guid is None:
            raise errors.ImageError_ArgsError(
                """You must supply a `guid` for
            the image. this is used for filename templating"""
            )

        # default to the resized images
        if selected_resizes is None:
            selected_resizes = list(resizerResultset.resized.keys())

        # quickly validate
        selected_resizes = self._validate__selected_resizes(
            resizerResultset, selected_resizes
        )

        # init our return dict
        filename_mapping = {}

        for size in selected_resizes:
            instructions = self._resizerConfig.resizesSchema[size]
            target_filename = size_to_filename(
                guid, size, resizerResultset, self.filename_template, instructions
            )

            # figure out the subdir
            subdir_name = self._saverConfig.subdir_public_name
            if "subdir_public" in instructions:
                subdir_name = instructions["subdir_public"]

            filename_mapping[size] = (target_filename, subdir_name)

        if check_archive_original(resizerResultset, archive_original=archive_original):
            filename_template_archive = self.filename_template_archive
            assert resizerResultset.original
            assert resizerResultset.original.format
            target_filename = filename_template_archive % {
                "guid": guid,
                "format": utils.PIL_type_to_standardized(
                    resizerResultset.original.format
                ),
            }
            subdir_name = self._saverConfig.subdir_archive_name
            filename_mapping["@archive"] = (target_filename, subdir_name)

        # return the filemapping
        return filename_mapping

    def files_save(  # type: ignore[override]
        self,
        resizerResultset: ResizerResultset,
        guid: str,
        selected_resizes: Optional[TYPE_selected_resizes] = None,
        archive_original: Optional[bool] = None,
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        """
        Returns a `dict` of resized images
        calls `self.register_image_file()`` if needed
        """
        if guid is None:
            raise errors.ImageError_ArgsError(
                """You must supply a `guid` for
            the image. this is used"""
            )

        # quickly validate
        selected_resizes = self._validate__selected_resizes(
            resizerResultset, selected_resizes
        )

        # and then we have the bucketed filenames...
        target_filenames = self.generate_filenames(
            resizerResultset,
            guid,
            selected_resizes=selected_resizes,
            archive_original=archive_original,
        )

        # log uploads for removal/tracking and return
        _saves = {}
        try:
            # and then we upload...
            for size in selected_resizes:
                (_filename, subdir_name) = target_filenames[size]
                target_dirname = os.path.join(self._saverConfig.filedir, subdir_name)
                if not os.path.exists(target_dirname):
                    os.makedirs(target_dirname)
                target_file = os.path.join(target_dirname, _filename)
                log.debug("Saving %s to %s " % (_filename, target_file))

                if not dry_run:
                    # upload
                    try:
                        with open(target_file, _io.FileWriteArgs) as _fh:
                            # resizerResultset.resized[size] == ResizedImage
                            # resizerResultset.resized[size].file == _io.BytesIO
                            _fh.write(resizerResultset.resized[size].file.getvalue())

                    except Exception as exc:
                        log.debug(
                            "encountered unexpected exception `%s` in `saver.localfile.SaverManager.files_save`",
                            exc,
                        )
                        raise

                    # log to external plugin too
                    if self._saverLogger:
                        self._saverLogger.log_save(
                            subdir_name=subdir_name,
                            filename=_filename,
                            file_size=resizerResultset.resized[size].file_size,
                            file_md5=resizerResultset.resized[size].file_md5,
                            filepath=target_file,
                        )

                # log for removal/tracking & return
                _saves[size] = (_filename, subdir_name)

            if "@archive" in target_filenames:
                size = "@archive"
                (_filename, subdir_name) = target_filenames[size]
                target_dirname = os.path.join(self._saverConfig.filedir, subdir_name)
                if not os.path.exists(target_dirname):
                    os.makedirs(target_dirname)
                target_file = os.path.join(target_dirname, _filename)
                log.debug("Saving %s to %s " % (_filename, target_file))

                if not dry_run:
                    # upload
                    with open(target_file, _io.FileWriteArgs) as _fh:
                        _fh.write(resizerResultset.original.file.getvalue())

                    # log to external plugin too
                    if self._saverLogger:
                        self._saverLogger.log_save(
                            subdir_name=subdir_name,
                            filename=_filename,
                            file_size=resizerResultset.original.file_size,
                            file_md5=resizerResultset.original.file_md5,
                            filepath=target_file,
                        )

                # log for removal/tracking & return
                _saves[size] = (_filename, subdir_name)

        except Exception as exc:
            # if we have ANY issues, we want to delete everything from amazon s3. otherwise this stuff is just hiding up there
            log.debug(
                "Error uploading... rolling back s3 items. unexpected exception `%s` in `saver.localfile.SaverManager.files_save`",
                exc,
            )
            _saves = self.files_delete(_saves)
            raise errors.ImageError_SaverUpload("error uploading")

        return _saves


class SaverSimpleAccess(_SaverCoreManager, _core.SaverSimpleAccess):
    def __init__(
        self,
        saverConfig=None,
        saverLogger=None,
        resizerConfig=None,
    ):
        self._saverConfig = saverConfig
        self._saverLogger = saverLogger
        self._resizerConfig = resizerConfig

    def file_save(  # type: ignore[override]
        self,
        subdir_name: str,
        filename: str,
        wrappedFile: BasicImage,
        dry_run: bool = False,
    ) -> TYPE_files_mapping:
        _saves = {}
        try:
            target_dirname = os.path.join(self._saverConfig.filedir, subdir_name)
            if not os.path.exists(target_dirname):
                os.makedirs(target_dirname)
            target_file = os.path.join(target_dirname, filename)
            log.debug("Saving %s to %s " % (filename, target_file))

            if not dry_run:
                # upload
                with open(target_file, _io.FileWriteArgs) as _fh:
                    _fh.write(wrappedFile.file.getvalue())

                # log to external plugin too
                if self._saverLogger:
                    self._saverLogger.log_save(
                        subdir_name=subdir_name,
                        filename=filename,
                        file_size=wrappedFile.file_size,
                        file_md5=wrappedFile.file_md5,
                        filepath=target_file,
                    )

            # log for removal/tracking & return
            _saves = self.simple_saves_mapping(subdir_name, filename)

            return _saves
        except Exception as exc:
            log.debug(
                "encountered unexpected exception `%s` in `saver.localfile.SaverSimpleAccess.files_save`",
                exc,
            )
            raise

    def simple_saves_mapping(
        self,
        subdir_name,
        filename,
    ) -> TYPE_files_mapping:
        _saves = {}
        _saves["%s||%s" % (subdir_name, filename)] = (filename, subdir_name)
        return _saves
